@@ -3,24 +3,33 @@
  *
  * Simple example of the draw cell, write cell routines
  *
- * $Id: draw.c,v 1.10 1999/07/24 01:49:10 lister Exp $
+ * $Id: draw.c,v 1.10 2002/08/20 16:21:09 dannybackx Exp $
  */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include <XbaeConfig.h>
+#endif
+#include <stdlib.h>
+#include <stdio.h>
+#ifdef USE_EDITRES
+#include <X11/Intrinsic.h>
+#include <X11/Xmu/Editres.h>
 #endif
 
-#include <Xbae/Matrix.h>
 #include <Xm/Text.h>
 #include <Xm/Form.h>
-#include <stdio.h>
-#if defined(HAVE_XPM_H)
-#include <xpm.h>
-#else  /* !HAVE_XPM_H */
-#ifdef HAVE_X11_XPM_H
-#include <X11/xpm.h>
-#endif /* HAVE_X11_XPM_H */
-#endif /* HAVE_XPM_H */
+#include <Xbae/Matrix.h>
+
+#if XmVersion >= 2001
+# define HAVE_XPM
+# include <Xm/XpmP.h>
+#elif defined(HAVE_XPM_H)
+# define HAVE_XPM
+# include <xpm.h>
+#elif defined(HAVE_X11_XPM_H)
+# define HAVE_XPM
+# include <X11/xpm.h>
+#endif
 
 #define ROWS 10000
 #define COLUMNS 20
@@ -31,10 +40,13 @@ static Pixmap pixmap = 0, mask;
 static Widget text;
 static int lastRow = -1;
 
-static void DrawCB(), WriteCB(), EnterCB(), LeaveCB();
-static void LoadMatrix();
+static void DrawCB(Widget w, XtPointer cd, XbaeMatrixDrawCellCallbackStruct *cbs);
+static void WriteCB(Widget w, XtPointer cd, XbaeMatrixWriteCellCallbackStruct *cbs);
+static void EnterCB(Widget w, XtPointer cd, XbaeMatrixEnterCellCallbackStruct *cbs);
+static void LeaveCB(Widget w, XtPointer cd, XbaeMatrixLeaveCellCallbackStruct *cbs);
+static void LoadMatrix(void);
 
-#if defined(HAVE_XPM_H) || defined(HAVE_X11_XPM_H)
+#ifdef HAVE_XPM
 /* XPM */
 static char * xbae_xpm[] =
 {
@@ -76,8 +88,22 @@ static unsigned char xbae_bits[] = {
    0x04, 0x78, 0x00, 0x00, 0x00, 0x00, 0x02, 0xf0, 0x00, 0x00, 0x00, 0x00
 };
 
+static String fallback[] = {
+	"Draw*mw.cellHighlightThickness:	2",
+	"Draw*mw.cellMarginWidth:		3",
+	"Draw*mw.cellMarginHeight:		5",
+	"Draw*mw.fixedRows:			2",
+	"Draw*mw.fixedColumns:			1",
+	"Draw*mw.traverseFixedCells:		True",
+	"Draw*mw.gridType:			grid_cell_shadow",
+	"Draw*mw.fontList:			-*-helvetica-medium-r-*-*-14-*-*-*-*-*-*-*",
+	"Draw*mw.allowColumnResize:		True",
+	"Draw*mw.selectScrollVisible:		False",
+	NULL
+};
+
 static void
-LoadMatrix()
+LoadMatrix(void)
 {
     int i, j;
 
@@ -87,21 +113,20 @@ LoadMatrix()
 }
 
 static void
-DrawCB( w, cd, cbs )
-Widget w;
-XtPointer cd;
-XbaeMatrixDrawCellCallbackStruct *cbs;
+DrawCB(Widget w, XtPointer cd, XbaeMatrixDrawCellCallbackStruct *cbs)
 {
     if( cbs->column == 2 )
     {
     if( !pixmap )
     {
-#if defined(HAVE_XPM_H) || defined(HAVE_X11_XPM_H)
+#ifdef HAVE_XPM
 	XpmAttributes attributes;
 
 	attributes.valuemask = 0;
 	attributes.valuemask |= XpmReturnInfos;
+#if XpmIncludeVersion > 30406
 	attributes.valuemask |= XpmReturnAllocPixels;
+#endif
 	attributes.valuemask |= XpmReturnExtensions;
 	
 	if( XpmCreatePixmapFromData(
@@ -133,7 +158,7 @@ XbaeMatrixDrawCellCallbackStruct *cbs;
 	   However, providing the values improves efficiency */
  	cbs->height = xbae_height;
  	cbs->width = xbae_width;
-#if defined(HAVE_XPM_H) || defined(HAVE_X11_XPM_H)
+#ifdef HAVE_XPM
 	cbs->depth = 8;
 #else
 	cbs->depth = 1;
@@ -148,10 +173,7 @@ XbaeMatrixDrawCellCallbackStruct *cbs;
 }
 
 static void
-WriteCB( w, cd, cbs )
-Widget w;
-XtPointer cd;
-XbaeMatrixWriteCellCallbackStruct *cbs;
+WriteCB(Widget w, XtPointer cd, XbaeMatrixWriteCellCallbackStruct *cbs)
 {
     char buf[ 128 ];
     
@@ -175,11 +197,9 @@ XbaeMatrixWriteCellCallbackStruct *cbs;
 }
 
 static void
-EnterCB( w, cd, cbs )
-Widget w;
-XtPointer cd;
-XbaeMatrixEnterCellCallbackStruct *cbs;
+EnterCB(Widget w, XtPointer cd, XbaeMatrixEnterCellCallbackStruct *cbs)
 {
+    printf("Enter\n");
     if( lastRow == cbs->row )
 	return;
 
@@ -192,25 +212,17 @@ XbaeMatrixEnterCellCallbackStruct *cbs;
 }
 
 static void
-LeaveCB( w, cd, cbs )
-Widget w;
-XtPointer cd;
-XbaeMatrixLeaveCellCallbackStruct *cbs;
+LeaveCB(Widget w, XtPointer cd, XbaeMatrixLeaveCellCallbackStruct *cbs)
 {
 	printf("Leave\n");
 }
 
 int
-main(argc, argv)
-int argc;
-char *argv[];
+main(int argc, char *argv[])
 {
     Widget toplevel, mw, form;
     XtAppContext app;
     int i;
-#ifdef USE_EDITRES
-    extern void _XEditResCheckMessages();
-#endif
     short widths[ COLUMNS ];
     char buf[ 128 ];
 
@@ -219,7 +231,7 @@ char *argv[];
     toplevel = XtVaAppInitialize(&app, "Draw",
 				 NULL, 0,
 				 &argc, argv,
-				 NULL,
+				 fallback,
 				 NULL);
 #ifdef USE_EDITRES
     XtAddEventHandler( toplevel, (EventMask)0, True,
@@ -244,10 +256,10 @@ char *argv[];
 	XmNrightAttachment, XmATTACH_FORM,
 	XmNtopAttachment, XmATTACH_FORM,
 	NULL);
-    XtAddCallback( mw, XmNdrawCellCallback, DrawCB, NULL );
-    XtAddCallback( mw, XmNwriteCellCallback, WriteCB, NULL );
-    XtAddCallback( mw, XmNenterCellCallback, EnterCB, NULL );
-    XtAddCallback( mw, XmNleaveCellCallback, LeaveCB, NULL );
+    XtAddCallback( mw, XmNdrawCellCallback,  (XtCallbackProc)DrawCB, NULL );
+    XtAddCallback( mw, XmNwriteCellCallback, (XtCallbackProc)WriteCB, NULL );
+    XtAddCallback( mw, XmNenterCellCallback, (XtCallbackProc)EnterCB, NULL );
+    XtAddCallback( mw, XmNleaveCellCallback, (XtCallbackProc)LeaveCB, NULL );
 
     sprintf( buf,
 	     "A matrix widget with %d rows and %d columns - That's %d cells!",
@@ -271,4 +283,3 @@ char *argv[];
     /*NOTREACHED*/
     return 0;
 }
-
