@@ -1,6 +1,7 @@
 /*
- * Copyright(c) 1992 Bell Communications Research, Inc. (Bellcore)
- * Copyright(c) 1995-99 Andrew Lister
+ * Copyright (c) 1992 Bell Communications Research, Inc. (Bellcore)
+ * Copyright (c) 1995-99 Andrew Lister
+ * Copyright  1999, 2000, 2001, 2002, 2003 by the LessTif Developers.
  *
  *                        All rights reserved
  * Permission to use, copy, modify and distribute this material for
@@ -21,7 +22,7 @@
  * LOST PROFITS OR OTHER INCIDENTAL OR CONSEQUENTIAL DAMAGES RELAT-
  * ING TO THE SOFTWARE.
  *
- * $Id: Macros.h,v 1.8 2000/03/28 21:44:08 danny Exp $
+ * $Id: Macros.h,v 1.41 2004/04/12 05:39:35 dannybackx Exp $
  */
 
 /*
@@ -31,58 +32,36 @@
 #ifndef _Xbae_Macros_h
 #define _Xbae_Macros_h
 
-#if defined(__cplusplus) || defined(c_plusplus)
-extern "C" {
-#endif
-
-#if XmVersion <= 1001
-#include <Xm/XmP.h>
-#else
 #include <Xm/DrawP.h>
-#endif
 #include <Xm/TextP.h>
 #include <Xm/PrimitiveP.h>
 
+
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /*
  * The third parameter is a string added for debugging purposes.
  */
-#if	0	/* Danny - DEBUG */
-#define DRAW_SHADOW(dpy, draw, dbg, tgc, bgc, sz, x, y, w, h, type)	\
-    { 									\
-	fprintf(stderr, "_XmDrawShadows(%s)(%d,%d,%d,%d,%d,%d)\n",	\
-		dbg, x, y, w, h, sz, type);				\
-	_XmDrawShadows(dpy, draw, tgc, bgc, x, y, w, h, sz, type);	\
-    }
+#if (XmVersion >= 2000)
+#define DRAW_SHADOW(dpy, draw, dbg, tgc, bgc, sz, x, y, w, h, type)		\
+	do {									\
+		DEBUGOUT(_XbaeDebug(__FILE__, NULL,				\
+			"XmeDrawShadows(%s) [%d,%d,%d,%d - %d %s]\n",		\
+			dbg, x, y, w, h, sz, _XbaeDebugShadowTypeToString(type)));	\
+		XmeDrawShadows(dpy, draw, tgc, bgc, x, y, w, h, sz, type);	\
+	} while(0)
 #else
-#if 1
-/*
- * BX (Builder Xcessory - see http://www.ics.com) is linked with a 1.2
- * version of Motif. Use of an Xbae widget which uses a Motif 2.* function
- * will crash BX when it tries to use the function.
- */
-#define DRAW_SHADOW(dpy, draw, dbg, tgc, bgc, sz, x, y, w, h, type)	\
-    _XmDrawShadows(dpy, draw, tgc, bgc, x, y, w, h, sz, type)
-
-#else
-#if XmVersion <= 1001
-#define DRAW_SHADOW(dpy, draw, dbg, tgc, bgc, sz, x, y, w, h, type)	\
-    _XmDrawShadow(dpy, draw, tgc, bgc, sz, x, y, w, h)
-#else
-#if XmVersion >= 2000
-#define DRAW_SHADOW(dpy, draw, dbg, tgc, bgc, sz, x, y, w, h, type)	\
-    XmeDrawShadows(dpy, draw, tgc, bgc, x, y, w, h, sz, type)
-#else
-#define DRAW_SHADOW(dpy, draw, dbg, tgc, bgc, sz, x, y, w, h, type)	\
-    _XmDrawShadows(dpy, draw, tgc, bgc, x, y, w, h, sz, type)
-#endif
-#endif
-#endif
+#define DRAW_SHADOW(dpy, draw, dbg, tgc, bgc, sz, x, y, w, h, type)		\
+	do {									\
+		DEBUGOUT(_XbaeDebug(__FILE__, NULL,				\
+			"_XmDrawShadows(%s) [%d,%d,%d,%d - %d %s]\n",		\
+			dbg, x, y, w, h, sz, _XbaeDebugShadowTypeToString(type)));	\
+		_XmDrawShadows(dpy, draw, tgc, bgc, x, y, w, h, sz, type);	\
+	} while(0)
 #endif
 
-#if XmVersion <= 1001
-#define DRAW_HIGHLIGHT(dpy, draw, tgc, x, y, w, h, sz, type) \
-	_XmDrawSimpleHighlight(dpy, draw, tgc, x, y, w, h, sz, type)
-#else
 #if XmVersion >= 2000
 #define DRAW_HIGHLIGHT(dpy, draw, tgc, x, y, w, h, sz, type) \
 	XmeDrawHighlight(dpy, draw, tgc, x, y, w, h, sz)
@@ -90,7 +69,6 @@ extern "C" {
 #define DRAW_HIGHLIGHT(dpy, draw, tgc, x, y, w, h, sz, type) \
 	_XmDrawHighlight(dpy, draw, tgc, x, y, w, h, sz, type)
 #endif
-#endif
 /*
  * Macros to retrieve our children.  Children must be created in this order.
  */
@@ -102,28 +80,35 @@ extern "C" {
 #define TextChild(mw)		(mw->composite.children[3])
 #define LeftClip(mw)		(mw->composite.children[4])
 #define RightClip(mw)		(mw->composite.children[5])
-#define TopClip(mw)		(mw->composite.children[6])
+#define TopClip(mw)		    (mw->composite.children[6])
 #define BottomClip(mw)		(mw->composite.children[7])
 
 #define VISIBLE_WIDTH(mw)	((int)ClipChild(mw)->core.width)
 #define VISIBLE_HEIGHT(mw)	((int)ClipChild(mw)->core.height)
 
-#define VERT_ORIGIN(mw)		(mw->matrix.top_row)
+/*
+ * SGO: changed the meaning of VERT_ORIGIN:
+ * now it contains the pixel position of the first row, same as
+ * HORIZ_ORIGIN, the top_row should now be accessed by TOP_ROW
+ */
+#define VERT_ORIGIN(mw)		(mw->matrix.vert_origin)
+#define TOP_ROW(mw)             (mw->matrix.top_row)
 #define TRAILING_VERT_ORIGIN(mw) (mw->matrix.rows - \
 				  (int)mw->matrix.trailing_fixed_rows)
 #define HORIZ_ORIGIN(mw)	(mw->matrix.horiz_origin)
 #define TRAILING_HORIZ_ORIGIN(mw) (mw->matrix.columns - \
 				   (int)mw->matrix.trailing_fixed_columns)
 
+#define	XbaeMultiLineCell(mw)	(mw->matrix.multi_line_cell)
+
 /*
  * Macros
  */
 
-/*
- * Max and Min are defined in Xm/XmP.h in 1.1, they are deprecated in 1.2
- */
-#if XmVersion >= 1002
+#ifndef Max
 #define Max(x, y)       (((x) > (y)) ? (x) : (y))
+#endif
+#ifndef Min
 #define Min(x, y)       (((x) < (y)) ? (x) : (y))
 #endif
 
@@ -141,13 +126,20 @@ extern "C" {
 				 mw->matrix.cell_highlight_thickness + \
 				 mw->matrix.text_shadow_thickness)
 
-#define SOME_ROW_HEIGHT(mw, row)	((mw->matrix.row_heights[row] * \
-				  FONT_HEIGHT(mw)) + \
-				 ((int)TEXT_HEIGHT_OFFSET(mw) * 2))
+#if 0 
+/* Tobias: use the positions array to calculate widths/heigts */
+/*
+ * This is - apparently - not good enough.
+ * #define SOME_ROW_HEIGHT(mw, row)	mw->matrix.row_heights[row]
+ */
+#define SOME_ROW_HEIGHT(mw, row)	(mw->matrix.row_heights ? mw->matrix.row_heights[row] : ROW_HEIGHT(mw))
 
-#define COLUMN_WIDTH(mw, col)	((mw->matrix.column_widths[col] * \
-				  FONT_WIDTH(mw)) + \
+#define COLUMN_WIDTH(mw, col)	((mw->matrix.column_width_in_pixels == True)? \
+				 mw->matrix.column_widths[col] :	\
+				 (mw->matrix.column_widths[col] * 	\
+				  FONT_WIDTH(mw)) + 			\
 				 ((int)TEXT_WIDTH_OFFSET(mw) * 2))
+#endif
 
 #define FONT_HEIGHT(mw)		(mw->matrix.font_height)
 
@@ -189,13 +181,14 @@ extern "C" {
 				 (mw->matrix.scrollbar_placement == \
 				  XmBOTTOM_LEFT)) ? VERT_SB_SPACE(mw) : 0)
 
+/* Tobias This is not symetrical with HORIZ_SB_POSITION */
 #define VERT_SB_POSITION(mw)	(((mw->matrix.scrollbar_placement == \
 				   XmTOP_LEFT) || \
 				  (mw->matrix.scrollbar_placement == \
 				   XmBOTTOM_LEFT)) ? 0 : \
 				 VertScrollChild(mw)->core.x - \
 				 mw->matrix.space)
-				  
+
 #define CLIP_HORIZ_VISIBLE_SPACE(mw) ((int)(VISIBLE_WIDTH(mw) +\
 				      FIXED_COLUMN_WIDTH(mw) + \
 				      TRAILING_FIXED_COLUMN_WIDTH(mw)))
@@ -204,34 +197,52 @@ extern "C" {
 				      FIXED_ROW_HEIGHT(mw) + \
 				      TRAILING_FIXED_ROW_HEIGHT(mw)))
 
+#if 0
+/* Tobias: These subtracted _SP_SPACE twice (because it can already be in _LABEL_OFFSET) */
 #define MATRIX_HORIZ_VISIBLE_SPACE(mw) ((int)(mw->core.width - \
-					 mw->manager.shadow_thickness - \
-					 COLUMN_LABEL_OFFSET(mw) -\
-					 VERT_SB_SPACE(mw)))
-
+                    mw->manager.shadow_thickness - \
+                    COLUMN_LABEL_OFFSET(mw) - \
+                    VERT_SB_SPACE(mw)))
 #define MATRIX_VERT_VISIBLE_SPACE(mw) ((int)(mw->core.height -\
-					mw->manager.shadow_thickness - \
-					ROW_LABEL_OFFSET(mw) -\
+                    mw->manager.shadow_thickness - \
+                    ROW_LABEL_OFFSET(mw) - \
+                    HORIZ_SB_SPACE(mw)))
+#else
+#define MATRIX_HORIZ_VISIBLE_SPACE(mw) ((int)(mw->core.width - \
+                    2 * mw->manager.shadow_thickness - \
+                    ROW_LABEL_WIDTH(mw) - \
+                    VERT_SB_SPACE(mw)))
+#define MATRIX_VERT_VISIBLE_SPACE(mw) ((int)(mw->core.height -\
+					2 * mw->manager.shadow_thickness - \
+					COLUMN_LABEL_HEIGHT(mw) -\
 					HORIZ_SB_SPACE(mw)))
+#endif
 
-#define NEED_HORIZ_FILL(mw)	(mw->matrix.fill && \
+#define NEED_HORIZ_FILL(mw)	(mw->matrix.fill && mw->matrix.horz_fill && \
 				  (MATRIX_HORIZ_VISIBLE_SPACE(mw) > \
 				    CLIP_HORIZ_VISIBLE_SPACE(mw)))
 
-#define NEED_VERT_FILL(mw)	(mw->matrix.fill && \
+#define NEED_VERT_FILL(mw)	(mw->matrix.fill && mw->matrix.vert_fill && \
 				  (MATRIX_VERT_VISIBLE_SPACE(mw) > \
 				    CLIP_VERT_VISIBLE_SPACE(mw)))
 
+#if 0
+/* Tobias: These are not symetrical and we probably want the others */
 #define	FILL_HORIZ_WIDTH(mw)	(mw->core.width - VISIBLE_WIDTH(mw) - \
 				  TRAILING_FIXED_COLUMN_WIDTH(mw) - \
 				  VERT_SB_SPACE(mw) - \
 				  FIXED_COLUMN_LABEL_OFFSET(mw) - \
 				  mw->manager.shadow_thickness)
-
 #define	FILL_VERT_HEIGHT(mw)	(mw->core.height - VISIBLE_HEIGHT(mw) - \
 				  TRAILING_FIXED_ROW_HEIGHT(mw) - \
 				  FIXED_ROW_LABEL_OFFSET(mw) - \
 				  mw->manager.shadow_thickness)
+#else
+#define	FILL_HORIZ_WIDTH(mw) \
+    (MATRIX_HORIZ_VISIBLE_SPACE(mw) - CLIP_HORIZ_VISIBLE_SPACE(mw))
+#define	FILL_VERT_HEIGHT(mw) \
+    (MATRIX_VERT_VISIBLE_SPACE(mw) - CLIP_VERT_VISIBLE_SPACE(mw))
+#endif
 
 #define SANITY_CHECK_ROW(mw, row) (row = (row >= mw->matrix.rows) ? \
 				      ((mw->matrix.rows > 0) ? \
@@ -259,86 +270,133 @@ extern "C" {
 
 #define LABEL_Y_OFFSET(mw)	(int)(mw->matrix.label_baseline)
 
-#define ROW_LABEL_WIDTH(mw)	((mw->matrix.row_labels \
-				 ? (mw->matrix.row_label_width * \
-				    LABEL_WIDTH(mw)) + \
-				 (int)TEXT_WIDTH_OFFSET(mw) * 2 : 0) + \
-				 (mw->matrix.button_labels ? \
-				 mw->matrix.cell_shadow_thickness * 2 : 0))
+#define ROW_LABEL_WIDTH(mw)	( \
+    (mw->matrix.row_labels \
+     ? (mw->matrix.row_label_width * LABEL_WIDTH(mw)) + \
+       (int)TEXT_WIDTH_OFFSET(mw) * 2 \
+     : 0) \
+    + (mw->matrix.button_labels \
+        ? mw->matrix.cell_shadow_thickness * 2  \
+        : 0))
 
-#define COLUMN_LABEL_HEIGHT(mw) (mw->matrix.column_labels \
-				 ? (int)TEXT_HEIGHT_OFFSET(mw) * 2 + \
-				 mw->matrix.column_label_maxlines * \
-				 LABEL_HEIGHT(mw) : 0)
+#define COLUMN_LABEL_HEIGHT(mw)		\
+    (mw->matrix.xmcolumn_labels		\
+      ? (int)TEXT_HEIGHT_OFFSET(mw) * 2 + LABEL_HEIGHT(mw) \
+      : (mw->matrix.column_labels \
+          ?	(int)TEXT_HEIGHT_OFFSET(mw) * 2 + mw->matrix.column_label_maxlines * LABEL_HEIGHT(mw) \
+          : 0))
 
+#if 1 /* Wether or not to check each column_position access for debugging */
 #define COLUMN_POSITION(mw, column) mw->matrix.column_positions[column]
 #define ROW_POSITION(mw, row) mw->matrix.row_positions[row]
- 
+#else
+#define COLUMN_POSITION(mw, column) xbaeCheckColumnPosition(mw, column)
+#define ROW_POSITION(mw, row) xbaeCheckRowPosition(mw, row)
+#endif
+
 #define FIXED_COLUMN_WIDTH(mw)	COLUMN_POSITION(mw, mw->matrix.fixed_columns)
 
+#if 0 /* Tobias: Somebody should have gone to bed earlier... */
+#define FIXED_ROW_WIDTH(mw)	ROW_POSITION(mw, mw->matrix.fixed_rows)
+#endif
+
+#if 0
+/* Tobias: use the positions array to calculate widths/heigts */
+#define	FIXED_ROW_HEIGHT(mw)	xbaeMatrixFixedRowHeight(mw)
+#define TRAILING_FIXED_ROW_HEIGHT(mw) xbaeMatrixTrailingFixedRowHeight(mw)
 #define TRAILING_FIXED_COLUMN_WIDTH(mw) \
                   (mw->matrix.trailing_fixed_columns ? \
 		    (COLUMN_POSITION(mw, mw->matrix.columns-1) + \
 		     COLUMN_WIDTH(mw, mw->matrix.columns-1) - \
 		     COLUMN_POSITION(mw, TRAILING_HORIZ_ORIGIN(mw))) : 0)
+#define NON_FIXED_TOTAL_WIDTH(mw) 	mw->matrix.non_fixed_total_width
+#define NON_FIXED_TOTAL_HEIGHT(mw) 	mw->matrix.non_fixed_total_height
+
+#else
+
+#define FIXED_ROW_HEIGHT(mw) ROW_POSITION(mw, mw->matrix.fixed_rows)
+
+#define SOME_ROW_HEIGHT(mw, row) \
+    (ROW_POSITION(mw, row + 1) - ROW_POSITION(mw, row))
+#define COLUMN_WIDTH(mw, column) \
+    (COLUMN_POSITION(mw, column + 1) - COLUMN_POSITION(mw, column))
+
+#define TRAILING_FIXED_ROW_HEIGHT(mw) \
+    (ROW_POSITION(mw, mw->matrix.rows) - ROW_POSITION(mw, TRAILING_VERT_ORIGIN(mw)))
+#define TRAILING_FIXED_COLUMN_WIDTH(mw) \
+    (COLUMN_POSITION(mw, mw->matrix.columns) - COLUMN_POSITION(mw, TRAILING_HORIZ_ORIGIN(mw)))
+
+#define NON_FIXED_TOTAL_HEIGHT(mw) \
+    (ROW_POSITION(mw, TRAILING_VERT_ORIGIN(mw)) - ROW_POSITION(mw, mw->matrix.fixed_rows))
+#define NON_FIXED_TOTAL_WIDTH(mw) \
+    (COLUMN_POSITION(mw, TRAILING_HORIZ_ORIGIN(mw)) - COLUMN_POSITION(mw, mw->matrix.fixed_columns))
+#endif
+
 
 #define COLUMN_LABEL_OFFSET(mw)	(ROW_LABEL_WIDTH(mw) + \
-				 VERT_SB_OFFSET(mw) + \
-				 (int)mw->manager.shadow_thickness)
-
-#define FIXED_COLUMN_LABEL_OFFSET(mw) (COLUMN_LABEL_OFFSET(mw) + \
-				       FIXED_COLUMN_WIDTH(mw))
-
-#define TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw) (FIXED_COLUMN_LABEL_OFFSET(mw)+\
-						VISIBLE_WIDTH(mw))
-
-#define FIXED_ROW_HEIGHT(mw)	((int)mw->matrix.fixed_rows * ROW_HEIGHT(mw))
-
-#define TRAILING_FIXED_ROW_HEIGHT(mw) ((int)mw->matrix.trailing_fixed_rows * \
-				       ROW_HEIGHT(mw))
-
+     VERT_SB_OFFSET(mw) + mw->manager.shadow_thickness)
 #define ROW_LABEL_OFFSET(mw)	(COLUMN_LABEL_HEIGHT(mw) + \
-				 HORIZ_SB_OFFSET(mw) + \
-				 mw->manager.shadow_thickness)
+     HORIZ_SB_OFFSET(mw) + mw->manager.shadow_thickness)
 
-#define FIXED_ROW_LABEL_OFFSET(mw) (ROW_LABEL_OFFSET(mw) + \
-				    FIXED_ROW_HEIGHT(mw))
+#define FIXED_COLUMN_LABEL_OFFSET(mw) \
+    (COLUMN_LABEL_OFFSET(mw) + FIXED_COLUMN_WIDTH(mw))
+#define FIXED_ROW_LABEL_OFFSET(mw) \
+    (ROW_LABEL_OFFSET(mw) + FIXED_ROW_HEIGHT(mw))
 
-#define UNATTACHED_TRAILING_COLUMNS_OFFSET(mw) (FIXED_COLUMN_LABEL_OFFSET(mw) + \
-					     VISIBLE_WIDTH(mw))
+#define UNATTACHED_TRAILING_COLUMNS_OFFSET(mw) \
+    (FIXED_COLUMN_LABEL_OFFSET(mw) + VISIBLE_WIDTH(mw))
+#define UNATTACHED_TRAILING_ROWS_OFFSET(mw) \
+    (FIXED_ROW_LABEL_OFFSET(mw)	+ VISIBLE_HEIGHT(mw))
 
-#define UNATTACHED_TRAILING_ROWS_OFFSET(mw) (FIXED_ROW_LABEL_OFFSET(mw) + \
-					     VISIBLE_HEIGHT(mw))
-
-#define ATTACHED_TRAILING_ROWS_OFFSET(mw) (mw->core.height - \
-					   mw->manager.shadow_thickness - \
-					   HORIZ_SB_SPACE(mw) - \
-					   TRAILING_FIXED_ROW_HEIGHT(mw))
+#define ATTACHED_TRAILING_ROWS_OFFSET(mw) ( \
+    mw->core.height\
+    - mw->manager.shadow_thickness \
+    - (HORIZ_SB_SPACE(mw) - HORIZ_SB_OFFSET(mw)) /* Tobias: 0 for XmTopX and HORIZ_SB_SPACE for XmBottomX */ \
+    - TRAILING_FIXED_ROW_HEIGHT(mw) \
+)
+#define ATTACHED_TRAILING_COLUMNS_OFFSET(mw) ( \
+    mw->core.width \
+    - mw->manager.shadow_thickness \
+    - (VERT_SB_SPACE(mw) - VERT_SB_OFFSET(mw)) /* Tobias: 0 for XmXLeft and VERT_SIB_SPACE for XmXRight */ \
+    - TRAILING_FIXED_COLUMN_WIDTH(mw) \
+)
 
 #define VERT_DEAD_SPACE_HEIGHT(mw)  ((int)(ATTACHED_TRAILING_ROWS_OFFSET(mw) - \
-				     UNATTACHED_TRAILING_ROWS_OFFSET(mw)))
+    UNATTACHED_TRAILING_ROWS_OFFSET(mw)))
+#define HORIZ_DEAD_SPACE_WIDTH(mw) ((int)(ATTACHED_TRAILING_COLUMNS_OFFSET(mw) - \
+    UNATTACHED_TRAILING_COLUMNS_OFFSET(mw)))
+
+#define NEED_VERT_DEAD_SPACE_FILL(mw) (HAS_ATTACHED_TRAILING_ROWS(mw) && \
+				      (VERT_DEAD_SPACE_HEIGHT(mw) > 0))
+#define NEED_HORIZ_DEAD_SPACE_FILL(mw) (HAS_ATTACHED_TRAILING_COLUMNS(mw) && \
+				       (HORIZ_DEAD_SPACE_WIDTH(mw) > 0))
 
 #define HAS_ATTACHED_TRAILING_ROWS(mw)  (mw->matrix.fill &&  \
 					 mw->matrix.trailing_attached_bottom && \
-					 mw->matrix.trailing_fixed_rows) 
-
+					 mw->matrix.trailing_fixed_rows)
 #define HAS_ATTACHED_TRAILING_COLUMNS(mw)  (mw->matrix.fill &&  \
 					 mw->matrix.trailing_attached_right && \
-					 mw->matrix.trailing_fixed_columns) 
+					 mw->matrix.trailing_fixed_columns)
+
+#if 0 /* Tobias: Make this symetrical with rows */
+#define TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw) (FIXED_COLUMN_LABEL_OFFSET(mw)+\
+						VISIBLE_WIDTH(mw))
+#else
+#define TRAILING_FIXED_COLUMN_LABEL_OFFSET(mw) (HAS_ATTACHED_TRAILING_COLUMNS(mw) ? \
+					     ATTACHED_TRAILING_COLUMNS_OFFSET(mw) : \
+					     UNATTACHED_TRAILING_COLUMNS_OFFSET(mw))
+#endif
 
 #define TRAILING_FIXED_ROW_LABEL_OFFSET(mw) (HAS_ATTACHED_TRAILING_ROWS(mw) ? \
 					     ATTACHED_TRAILING_ROWS_OFFSET(mw) : \
 					     UNATTACHED_TRAILING_ROWS_OFFSET(mw))
 
-#define NEED_VERT_DEAD_SPACE_FILL(mw) (HAS_ATTACHED_TRAILING_ROWS(mw) && \
-				      (VERT_DEAD_SPACE_HEIGHT(mw) > 0))
-
-#define NON_FIXED_TOTAL_WIDTH(mw) 	mw->matrix.non_fixed_total_width
-
+#if 0 /* Wrong and unused */
 #define CELL_TOTAL_HEIGHT(mw)	((mw->matrix.rows - \
 				  (int) mw->matrix.fixed_rows - \
 				  (int) mw->matrix.trailing_fixed_rows) \
 				 * ROW_HEIGHT(mw))
+#endif
 
 #define IS_LEADING_FIXED_COLUMN(mw, column) (column < (int)mw->matrix.fixed_columns)
 
@@ -357,11 +415,11 @@ extern "C" {
 #define IS_FIXED(mw, row, column) (IS_FIXED_ROW(mw, row) || \
 				   IS_FIXED_COLUMN(mw, column))
 
-#define IS_CLIPPED(mw, row, column) ((row >= (int)mw->matrix.fixed_rows) && \
-				       (column >= \
-					(int)mw->matrix.fixed_columns) && \
-				       (row < TRAILING_VERT_ORIGIN(mw)) && \
-				       (column < TRAILING_HORIZ_ORIGIN(mw)))
+#define IS_CLIPPED(mw, row, column) ( \
+    (row    >= (int)mw->matrix.fixed_rows) && \
+    (column >= (int)mw->matrix.fixed_columns) && \
+    (row    < TRAILING_VERT_ORIGIN(mw)) && \
+    (column < TRAILING_HORIZ_ORIGIN(mw)))
 
 /*
  * I hereby dub thee, Jay Schmidgall, as Sir Silly Macro - writer of
@@ -375,7 +433,7 @@ extern "C" {
  * actual code. I did the vertical scrollbar stuff first, then the
  * horizontal. For the horizontal, I just referenced the other
  * orientation, and it worked. */
-    
+
 /* If we have fixed rows, the y position of the vsb must
  * be offset starting from the fixed rows. Otherwise, we
  * want it to be even with the matrix area top.  Perhaps
@@ -388,14 +446,14 @@ extern "C" {
 			     HORIZ_SB_HEIGHT(mw) : 0) : \
 			    COLUMN_LABEL_HEIGHT(mw) + \
 			    HORIZ_SB_OFFSET(mw))
-	
+
 #define HSB_X_POSITION(mw) (mw->matrix.fixed_columns ? \
 			    FIXED_COLUMN_LABEL_OFFSET(mw) - \
 			    VERT_SB_OFFSET(mw) + \
 			    ((scrollbar_left && has_vert) ? \
 			     VERT_SB_WIDTH(mw) : 0) : \
 			    ROW_LABEL_WIDTH(mw) + VERT_SB_OFFSET(mw))
-	
+
 /* I started trying to write a comment that sort of explains this macro.
  * Then I quit. What the heck, it works. Feel free on your own to make
  * up something that reassures you. I myself have the feeling that there
@@ -440,110 +498,77 @@ extern "C" {
 			 VERT_SB_OFFSET(mw) - \
 			 (has_vert ? VERT_SB_WIDTH(mw) : 0)))
 
+/* SGO: This macro determines the correct size of the vsb-slider
+#define VSB_SLIDER_SIZE(mw)   (CLIP_VERT_VISIBLE_SPACE(mw) < 1 ? 1 : \
+			      (CLIP_VERT_VISIBLE_SPACE(mw) > xbaeMaxVertScroll(mw) ?  \
+			       xbaeMaxVertScroll(mw) : CLIP_VERT_VISIBLE_SPACE(mw)))
+#define VSB_SLIDER_SIZE(mw)   (XtHeight(ClipChild(mw)))
+*/
+#define	VSB_SLIDER_SIZE(mw)	((XtHeight(ClipChild(mw)) < xbaeMaxVertScroll(mw)) ? \
+				XtHeight(ClipChild(mw)) : xbaeMaxVertScroll(mw))
+
+#if 0 /* unused */
 #define CELL_WINDOW(mw, row, column) \
 	(IS_FIXED(mw, row, column) ? XtWindow(mw) : XtWindow(ClipChild(mw)))
+#endif
 
 /* Inline functions */
-#define xbaeFreeColumnWidths(mw)	{ \
-					       if (mw->matrix.column_widths) \
-					       XtFree((XtPointer) \
-					       mw->matrix.column_widths); \
-					}
 
-#define xbaeFreeRowHeights(mw)	{ \
-					       if (mw->matrix.row_heights) \
-					       XtFree((XtPointer) \
-					       mw->matrix.row_heights); \
-					}
+#define xbaeFreeArray(array) \
+    do{ \
+        if(array) XtFree((XtPointer) array); \
+    } while(0)
 
-#define xbaeFreeColumnMaxLengths(mw)	{ \
-					       if (mw->matrix.\
-					       column_max_lengths) \
-					       XtFree((XtPointer) \
-					       mw->matrix. \
-					       column_max_lengths); \
-					}
+#define xbaeFreeColumnWidths(mw) \
+    xbaeFreeArray(mw->matrix.column_widths)
 
-#define xbaeFreeColumnPositions(mw)		{ \
-					       if (mw->matrix. \
-					       column_positions) \
-					       XtFree((XtPointer) \
-					       mw->matrix. \
-					       column_positions); \
-					}
+#define xbaeFreeRowHeights(mw) \
+    xbaeFreeArray(mw->matrix.row_heights)
 
-#define xbaeFreeColumnAlignments(mw)	{ \
-					       if (mw->matrix. \
-					       column_alignments) \
-					       XtFree((XtPointer) \
-					       mw->matrix. \
-					       column_alignments); \
-					}
+#define xbaeFreeColumnMaxLengths(mw) \
+    xbaeFreeArray(mw->matrix.column_max_lengths)
+    
+#define xbaeFreeColumnPositions(mw) \
+    xbaeFreeArray(mw->matrix.column_positions)
 
-#define xbaeFreeColumnButtonLabels(mw)	{ \
-					       if (mw->matrix. \
-					       column_button_labels) \
-					       XtFree((XtPointer) \
-					       mw->matrix. \
-					       column_button_labels); \
-					}
+#define xbaeFreeRowPositions(mw) \
+    xbaeFreeArray(mw->matrix.row_positions)
+    
+#define xbaeFreeColumnAlignments(mw) \
+    xbaeFreeArray(mw->matrix.column_alignments)
 
-#define xbaeFreeRowButtonLabels(mw)	{ \
-					       if (mw->matrix. \
-					       row_button_labels) \
-					       XtFree((XtPointer) \
-					       mw->matrix. \
-					       row_button_labels); \
-					}
+#define xbaeFreeColumnButtonLabels(mw) \
+    xbaeFreeArray(mw->matrix.column_button_labels)
 
-#define xbaeFreeColumnLabelAlignments(mw)	{ \
-					       if (mw->matrix. \
-					       column_label_alignments) \
-					       XtFree((XtPointer) \
-					       mw->matrix. \
-					       column_label_alignments); \
-					}
+#define xbaeFreeShowColumnArrows(mw) \
+    xbaeFreeArray(mw->matrix.show_column_arrows)
 
-#define xbaeFreeRowUserData(mw)		{ \
-					       if (mw->matrix.\
-					       row_user_data) \
-					       XtFree((XtPointer) \
-					       mw->matrix.\
-					       row_user_data); \
-					}
+#define xbaeFreeColumnFontBold(mw) \
+    xbaeFreeArray(mw->matrix.column_font_bold)
 
-#define xbaeFreeColumnUserData(mw)	{ \
-					       if (mw->matrix.\
-					       column_user_data) \
-					       XtFree((XtPointer) \
-					       mw->matrix.\
-					       column_user_data); \
-					}
+#define xbaeFreeRowButtonLabels(mw) \
+    xbaeFreeArray(mw->matrix.row_button_labels)
 
-#define xbaeFreeRowShadowTypes(mw)	{ \
-					       if (mw->matrix.\
-					       row_shadow_types) \
-					       XtFree((XtPointer) \
-					       mw->matrix.\
-					       row_shadow_types); \
-					}
+#define xbaeFreeColumnLabelAlignments(mw) \
+    xbaeFreeArray(mw->matrix.column_label_alignments)
 
-#define xbaeFreeColumnShadowTypes(mw)	{ \
-					       if (mw->matrix.\
-					       column_shadow_types) \
-					       XtFree((XtPointer) \
-					       mw->matrix.\
-					       column_shadow_types); \
-					}
+#define xbaeFreeRowUserData(mw) \
+    xbaeFreeArray(mw->matrix.row_user_data)
+
+#define xbaeFreeColumnUserData(mw) \
+    xbaeFreeArray(mw->matrix.column_user_data)
+
+#define xbaeFreeRowShadowTypes(mw) \
+    xbaeFreeArray(mw->matrix.row_shadow_types)
+    
+#define xbaeFreeColumnShadowTypes(mw) \
+    xbaeFreeArray(mw->matrix.column_shadow_types)
 
 #define CreateRowPositions(mw) \
 	(int *)XtMalloc((mw->matrix.rows+1) * sizeof(int))
 #define CreateColumnPositions(mw) \
 	(int *)XtMalloc((mw->matrix.columns+1) * sizeof(int))
 
-#define YtoRow(mw, y)			((y) / ROW_HEIGHT(mw))
-
-
 /*
  * Evaluates to 1 if two Rectangles overlap, 0 if no overlap
  */
@@ -556,14 +581,23 @@ extern "C" {
  * Intersect rectangles r1 and r2, place the result in res.
  * Result will be in r1's coord system.
  */
-#define X_INTERSECT(r1, r2, res) { (res).x1 = Max((r1).x1, (r2).x1) - (r1).x1;\
-    (res).x2 = Min((r1).x2, (r2).x2) - (r1).x1;}
+#define X_INTERSECT(r1, r2, res) \
+    do { \
+        (res).x1 = Max((r1).x1, (r2).x1) - (r1).x1; \
+        (res).x2 = Min((r1).x2, (r2).x2) - (r1).x1; \
+    } while(0)
 
-#define Y_INTERSECT(r1, r2, res) { (res).y1 = Max((r1).y1, (r2).y1) - (r1).y1;\
-    (res).y2 = Min((r1).y2, (r2).y2) - (r1).y1;}
+#define Y_INTERSECT(r1, r2, res) \
+    do { \
+        (res).y1 = Max((r1).y1, (r2).y1) - (r1).y1; \
+        (res).y2 = Min((r1).y2, (r2).y2) - (r1).y1; \
+    } while(0)
 
-#define INTERSECT(r1, r2, res)	{ X_INTERSECT(r1, r2, res); \
-    Y_INTERSECT(r1, r2, res); }
+#define INTERSECT(r1, r2, res) \
+    do { \
+        X_INTERSECT(r1, r2, res); \
+        Y_INTERSECT(r1, r2, res); \
+    } while(0)
 
 /*
  * Evaluates to 1 if the point is in the Rectangle, 0 if not
@@ -581,18 +615,17 @@ extern "C" {
 /*
  * Set a Rectangle. (x1,y1) is upper left corner, (x2,y2) is lower right corner
  */
-#define SETRECT(r, X1, Y1, X2, Y2) { (r).x1 = X1; (r).y1 = Y1; \
-    (r).x2 = X2; (r).y2 = Y2; }
+#define SETRECT(r, X1, Y1, X2, Y2) \
+    do{ \
+        (r).x1 = X1; (r).y1 = Y1; \
+        (r).x2 = X2; (r).y2 = Y2; \
+    } while(0)
 
 #ifdef NEED_WCHAR
 #define TWO_BYTE_FONT(mw)	(mw->matrix.font->max_byte1 != 0)
 #endif
 
-#ifdef NEED_24BIT_VISUAL
-#define GC_PARENT_WINDOW(w)	XtWindow(get_shell_ancestor((Widget)w))
-#else
-#define GC_PARENT_WINDOW(w)	RootWindowOfScreen(XtScreen(w))
-#endif
+#define GC_PARENT_WINDOW(w)	XtWindow(_XbaeGetShellAncestor((Widget)w))
 
 /*
  * End of array flags for the array type converters
@@ -601,29 +634,32 @@ extern "C" {
 #define BAD_HEIGHT	-1
 #define BAD_MAXLENGTH	0
 #define BAD_ALIGNMENT	3	/* see Xm.h */
-#define BAD_PIXEL	0x10000000 /* normally 256 indices */ 
-
+#define BAD_PIXEL	0x10000000 /* normally 256 indices */
 
 /*
  * SetClipMask flags for indicating clip areas
  */
-#define CLIP_NONE			0x0000
-#define CLIP_FIXED_COLUMNS		0x0001
-#define CLIP_FIXED_ROWS			0x0002
+#define CLIP_NONE                   0x0000
+#define CLIP_FIXED_COLUMNS          0x0001
+#define CLIP_FIXED_ROWS             0x0002
 #define CLIP_TRAILING_FIXED_COLUMNS	0x0004
 #define CLIP_TRAILING_FIXED_ROWS	0x0008
 #define CLIP_BETWEEN_FIXED_ROWS		0x0010
-#define CLIP_VISIBLE_HEIGHT		0x0020
+#define CLIP_VISIBLE_HEIGHT         0x0020
 #define CLIP_TRAILING_HORIZ_FILL	0x0040
 #define CLIP_TRAILING_VERT_FILL		0x0080
 #define CLIP_COLUMN_LABELS     		0x0100
 #define CLIP_ROW_LABELS     		0x0200
-#define CLIP_ALL	     		0x0400
+#if 0 /* Tobias: Unused */
+#define CLIP_ALL	     		    0x0400
+#endif
+#define CLIP_BETWEEN_FIXED_COLUMNS  0x0800
+#define CLIP_VISIBLE_WIDTH          0x1000
 
 /*
  * Row and Column grid shadow/line redraw reasons
  */
-#define GRID_REDRAW_EXPOSE		0x0000
+#define GRID_REDRAW_EXPOSE          0x0000
 #define GRID_REDRAW_SCROLL_VERT		0x0001
 #define GRID_REDRAW_SCROLL_HORIZ	0x0002
 #define GRID_REDRAW_EDIT		(GRID_REDRAW_SCROLL_VERT | \
@@ -631,11 +667,9 @@ extern "C" {
 /*
  * Un/Highlight indicators
  */
-#if XmVersion >= 1002
 #define GRID_REDRAW_HIGHLIGHT		0x0100
 #define HIGHLIGHTING_SOMETHING		0x0F
 #define UNHIGHLIGHTING_SOMETHING	0xF0
-#endif
 
 /*
  * Grid shadow/line detectors
@@ -647,8 +681,9 @@ extern "C" {
 #define IN_GRID_ROW_MODE(mw)	(mw->matrix.grid_type & GRID_MODE_ROW)
 #define IN_GRID_COLUMN_MODE(mw)	(mw->matrix.grid_type & GRID_MODE_COLUMN)
 
-#if defined(__cplusplus) || defined(c_plusplus)
+#ifdef __cplusplus
 }
 #endif
 
+
 #endif /* _Xbae_Macros_h */
