@@ -1,6 +1,7 @@
 /*
  * Copyright(c) 1992 Bell Communications Research, Inc. (Bellcore)
  * Copyright(c) 1999 Andrew Lister
+ * Copyright  1999,2000,2001,2002 by the LessTif Developers.
  *
  *                        All rights reserved
  * Permission to use, copy, modify and distribute this material for
@@ -23,7 +24,7 @@
  *
  * CaptionWidget Author: Andrew Wason, Bellcore, aw@bae.bellcore.com
  * 
- * $Id: Caption.c,v 1.4 2000/02/29 19:29:24 danny Exp $
+ * $Id: Caption.c,v 1.12 2002/07/14 09:34:09 dannybackx Exp $
  */
 
 /*
@@ -35,14 +36,20 @@
 #endif
 
 #include <ctype.h>
+#include <string.h>
+
 #include <X11/StringDefs.h>
+
 #include <Xm/XmP.h>
 #include <Xm/Label.h>
+
 #include <Xbae/Macros.h>
 #include <Xbae/CaptionP.h>
 
-#ifndef tolower
-#define tolower(c)      ((c) - 'A' + 'a')
+#include <XbaeDebug.h>
+
+#ifdef	WITH_DMALLOC
+#include <dmalloc.h>
 #endif
 
 #define OffsetOf(field)	XtOffsetOf(XbaeCaptionRec, field)
@@ -481,7 +488,7 @@ GetValuesHook(Widget w,
      * We don't save a copy of the label_string or font_list.
      * If the user wants these, we get them from the label widget.
      */
-    for (i = 0; i < *num_args; i++)
+    for (i = 0; i < (int)*num_args; i++)
 	if (strcmp(args[i].name, XmNlabelString) == 0)
 	    XtGetValues(LabelChild(cw), &args[i], 1);
 	else if (strcmp(args[i].name, XmNfontList) == 0)
