@@ -1,6 +1,8 @@
 /*
  * Copyright(c) 1992 Bell Communications Research, Inc. (Bellcore)
  * Copyright(c) 1995-99 Andrew Lister
+ * Copyright  1999,2000,2001,2002 by the LessTif Developers.
+ *
  *                        All rights reserved
  * Permission to use, copy, modify and distribute this material for
  * any purpose and without fee is hereby granted, provided that the
@@ -22,7 +24,7 @@
  *
  * MatrixWidget Author: Andrew Wason, Bellcore, aw@bae.bellcore.com
  *
- * $Id: Converters.c,v 1.2 2000/02/26 15:52:36 danny Exp $
+ * $Id: Converters.c,v 1.11 2003/10/20 19:16:48 dannybackx Exp $
  */
 
 /*
@@ -41,22 +43,24 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 /* #include <X11/Xos.h> */
 #include <X11/StringDefs.h>
+
 #include <Xm/XmP.h>
 #include <Xm/AtomMgr.h>
+
 #include <Xbae/MatrixP.h>
 #include <Xbae/Converters.h>
 
-#ifndef tolower
-#define tolower(c)      ((c) - 'A' + 'a')
-#endif
+#include <XbaeDebug.h>
 
-static Boolean StringsAreEqual(String, String, int);
+
+static Boolean XbaeStringsAreEqual(String, String, int);
 
 /* ARGSUSED */
 Boolean
-CvtStringToStringArray(Display *dpy,
+XbaeCvtStringToStringArray(Display *dpy,
 		XrmValuePtr args,
 		Cardinal *num_args,
 		XrmValuePtr from,
@@ -207,7 +211,7 @@ CvtStringToStringArray(Display *dpy,
  */
 /* ARGSUSED */
 void
-StringArrayDestructor(XtAppContext app,
+XbaeStringArrayDestructor(XtAppContext app,
 			XrmValuePtr to,
 			XtPointer converter_data,
 			XrmValuePtr args,
@@ -216,18 +220,20 @@ StringArrayDestructor(XtAppContext app,
     String *array = *(String **) to->addr;
     String *entry;
 
+    fprintf(stderr, "XbaeStringArrayDestructor(%p)\n", array);
+
     if (array == NULL)
 	return;
 
     for (entry = array; *entry != NULL; entry++)
-	XtFree((XtPointer) * entry);
+	XtFree((char *) * entry);
 
-    XtFree((XtPointer) array);
+    XtFree((char *) array);
 }
 
 /* ARGSUSED */
 Boolean
-CvtStringToCellTable(Display *dpy,
+XbaeCvtStringToCellTable(Display *dpy,
 		XrmValuePtr args,
 		Cardinal *num_args,
 		XrmValuePtr from,
@@ -289,14 +295,14 @@ CvtStringToCellTable(Display *dpy,
             lto.addr = (char *)(&array[k]);
             lto.size = sizeof(String *);
             
-            if (!CvtStringToStringArray(dpy, args, num_args, &lfrom, &lto, data))
+            if (!XbaeCvtStringToStringArray(dpy, args, num_args, &lfrom, &lto, data))
             {
                 *ch = c;
 		XtDisplayStringConversionWarning(dpy, from->addr,
 						 XmRCellTable);
 		array[k] = (String *)0;
                 lto.addr = (char *)&array;
-                StringCellDestructor(0, &lto, 0, 0, 0);
+                XbaeStringCellDestructor(0, &lto, 0, 0, 0);
        		return False;
 	    }
             *ch = c;
@@ -323,7 +329,7 @@ CvtStringToCellTable(Display *dpy,
  */
 /* ARGSUSED */
 void
-StringCellDestructor(XtAppContext app,
+XbaeStringCellDestructor(XtAppContext app,
 		XrmValuePtr to,
 		XtPointer converter_data,
 		XrmValuePtr args,
@@ -332,17 +338,19 @@ StringCellDestructor(XtAppContext app,
     String **array = *(String ***) to->addr;
     String **entry, *row;
 
+    fprintf(stderr, "XbaeStringCellDestructor(%p)\n", array);
+
     if (array == NULL)
 	return;
 
     for (entry = array; *entry != NULL; entry++)
     {
         for (row = *entry; *row != NULL; row++)
-	    XtFree((XtPointer) *row);
-	XtFree((XtPointer) *entry);
+	    XtFree((char *) *row);
+	XtFree((char *) *entry);
     }
 
-    XtFree((XtPointer) array);
+    XtFree((char *) array);
 }
 
 /*
@@ -351,7 +359,7 @@ StringCellDestructor(XtAppContext app,
  */
 /* ARGSUSED */
 Boolean
-CvtStringToWidthArray(Display *dpy,
+XbaeCvtStringToWidthArray(Display *dpy,
 			XrmValuePtr args,
 			Cardinal *num_args,
 			XrmValuePtr from,
@@ -429,7 +437,7 @@ CvtStringToWidthArray(Display *dpy,
  */
 /* ARGSUSED */
 void
-WidthArrayDestructor(XtAppContext app,
+XbaeWidthArrayDestructor(XtAppContext app,
 		XrmValuePtr to,
 		XtPointer converter_data,
 		XrmValuePtr args,
@@ -437,7 +445,7 @@ WidthArrayDestructor(XtAppContext app,
 {
     short *array = *(short **) to->addr;
 
-    XtFree((XtPointer) array);
+    XtFree((char *) array);
 }
 
 /*
@@ -446,7 +454,7 @@ WidthArrayDestructor(XtAppContext app,
  */
 /* ARGSUSED */
 Boolean
-CvtStringToMaxLengthArray(Display *dpy,
+XbaeCvtStringToMaxLengthArray(Display *dpy,
 			XrmValuePtr args,
 			Cardinal *num_args,
 			XrmValuePtr from,
@@ -525,7 +533,7 @@ CvtStringToMaxLengthArray(Display *dpy,
  */
 /* ARGSUSED */
 void
-MaxLengthArrayDestructor(XtAppContext app,
+XbaeMaxLengthArrayDestructor(XtAppContext app,
 			XrmValuePtr to,
 			XtPointer converter_data,
 			XrmValuePtr args,
@@ -533,7 +541,7 @@ MaxLengthArrayDestructor(XtAppContext app,
 {
     int *array = *(int **) to->addr;
 
-    XtFree((XtPointer) array);
+    XtFree((char *) array);
 }
 
 /*
@@ -542,7 +550,7 @@ MaxLengthArrayDestructor(XtAppContext app,
  * Used by StringToAlignmentArray converter.
  */
 static Boolean
-StringsAreEqual(String in, String test, int length)
+XbaeStringsAreEqual(String in, String test, int length)
 {
     int i;
 
@@ -576,7 +584,7 @@ StringsAreEqual(String in, String test, int length)
  */
 /* ARGSUSED */
 Boolean
-CvtStringToBooleanArray(Display *dpy,
+XbaeCvtStringToBooleanArray(Display *dpy,
 			XrmValuePtr args,
 			Cardinal *num_args,
 			XrmValuePtr from,
@@ -634,13 +642,13 @@ CvtStringToBooleanArray(Display *dpy,
 	    while (isspace(*start))
 		start++;
 
-	    if (StringsAreEqual(start, "true", 4))
+	    if (XbaeStringsAreEqual(start, "true", 4))
 		array[i] = True;
-	    else if (StringsAreEqual(start, "1", 1))
+	    else if (XbaeStringsAreEqual(start, "1", 1))
 		array[i] = True;
-	    else if (StringsAreEqual(start, "false", 5))
+	    else if (XbaeStringsAreEqual(start, "false", 5))
 		array[i] = False;
-	    else if (StringsAreEqual(start, "0", 1))
+	    else if (XbaeStringsAreEqual(start, "0", 1))
 		array[i] = False;
 	    else
 	    {
@@ -674,7 +682,7 @@ CvtStringToBooleanArray(Display *dpy,
  */
 /* ARGSUSED */
 void
-BooleanArrayDestructor(XtAppContext app,
+XbaeBooleanArrayDestructor(XtAppContext app,
 			XrmValuePtr to,
 			XtPointer converter_data,
 			XrmValuePtr args,
@@ -682,7 +690,7 @@ BooleanArrayDestructor(XtAppContext app,
 {
     Boolean *array = *(Boolean **) to->addr;
 
-    XtFree((XtPointer) array);
+    XtFree((char *) array);
 }
 
 
@@ -692,7 +700,7 @@ BooleanArrayDestructor(XtAppContext app,
  */
 /* ARGSUSED */
 Boolean
-CvtStringToAlignmentArray(Display *dpy,
+XbaeCvtStringToAlignmentArray(Display *dpy,
 			XrmValuePtr args,
 			Cardinal *num_args,
 			XrmValuePtr from,
@@ -757,11 +765,11 @@ CvtStringToAlignmentArray(Display *dpy,
 	    while (isspace(*start))
 		start++;
 
-	    if (StringsAreEqual(start, "alignment_beginning", 19))
+	    if (XbaeStringsAreEqual(start, "alignment_beginning", 19))
 		array[i] = XmALIGNMENT_BEGINNING;
-	    else if (StringsAreEqual(start, "alignment_center", 16))
+	    else if (XbaeStringsAreEqual(start, "alignment_center", 16))
 		array[i] = XmALIGNMENT_CENTER;
-	    else if (StringsAreEqual(start, "alignment_end", 13))
+	    else if (XbaeStringsAreEqual(start, "alignment_end", 13))
 		array[i] = XmALIGNMENT_END;
 	    else
 	    {
@@ -795,7 +803,7 @@ CvtStringToAlignmentArray(Display *dpy,
  */
 /* ARGSUSED */
 void
-AlignmentArrayDestructor(XtAppContext app,
+XbaeAlignmentArrayDestructor(XtAppContext app,
 			XrmValuePtr to,
 			XtPointer converter_data,
 			XrmValuePtr args,
@@ -803,7 +811,129 @@ AlignmentArrayDestructor(XtAppContext app,
 {
     unsigned char *array = *(unsigned char **) to->addr;
 
-    XtFree((XtPointer) array);
+    XtFree((char *) array);
+}
+
+/*
+ * Convert a comma separated list of shadow types to array of shadowTypes
+ * (unsigned chars).  The array is terminated by XmNO_LINE.
+ */
+/* ARGSUSED */
+Boolean
+XbaeCvtStringToShadowTypeArray(Display *dpy,
+			XrmValuePtr args,
+			Cardinal *num_args,
+			XrmValuePtr from,
+			XrmValuePtr to,
+			XtPointer *data)
+{
+    static unsigned char *array;
+    String start = from->addr;
+    char *ch;
+    int i, count;
+
+    if (*num_args != 0)
+	XtAppWarningMsg(
+	    XtDisplayToApplicationContext(dpy),
+	    "cvtStringToShadowTypeArray", "wrongParameters",
+	    "XbaeMatrix",
+	    "String to ShadowTypeArray conversion needs no extra arguments",
+	    NULL, NULL);
+
+    /*
+     * User didn't provide enough space
+     */
+    if (to->addr != NULL && to->size < sizeof(unsigned char *))
+    {
+	to->size = sizeof(unsigned char *);
+
+	return False;
+    }
+
+    if (start == NULL || *start == '\0')
+	array = NULL;
+
+    else
+    {
+
+	/*
+	 * Count the comma separated alignments
+	 */
+	for (ch = start, count = 1; *ch != '\0'; ch++)
+	    if (*ch == ',')
+		count++;
+
+	/*
+	 * Malloc the array
+	 */
+	array = (unsigned char *) XtMalloc((count + 1) * sizeof(unsigned char));
+
+	array[count] = XmNO_LINE;
+
+	/*
+	 * Compare each substring to the alignment strings.
+	 * If we find a bad one, display a warning and fail.
+	 * We should be able to use XtCallConverter on _XmCvtStringToAlignment,
+	 * but that function is static so we have to duplicate its
+	 * functionality.
+	 */
+	for (i = 0; i < count; i++)
+	{
+	    /*
+	     * Skip leading white space
+	     */
+	    while (isspace(*start))
+		start++;
+
+	    if (XbaeStringsAreEqual(start, "shadow_in", 9))
+		array[i] = XmSHADOW_IN;
+	    else if (XbaeStringsAreEqual(start, "shadow_out", 10))
+		array[i] = XmSHADOW_OUT;
+#if 0
+	    else if (XbaeStringsAreEqual(start, "shadow_out", 10))
+		array[i] = XmSHADOW_OUT;
+#endif
+	    else
+	    {
+		XtDisplayStringConversionWarning(dpy, from->addr,
+						 XmRShadowTypeArray);
+		XtFree((void*)array);
+		return False;
+	    }
+
+	    /*
+	     * Find the comma at the end of this alignment
+	     */
+	    /* EMPTY */
+	    for (; *start != '\0' && *start != ','; start++);
+	    start++;
+	}
+    }
+
+    if (to->addr == NULL)
+	to->addr = (XtPointer) & array;
+    else
+	*(unsigned char **) to->addr = array;
+    to->size = sizeof(unsigned char *);
+
+    return True;
+}
+
+
+/*
+ * Free the alignment array allocated by the String to ShadowTypeArray converter
+ */
+/* ARGSUSED */
+void
+XbaeShadowTypeArrayDestructor(XtAppContext app,
+			XrmValuePtr to,
+			XtPointer converter_data,
+			XrmValuePtr args,
+			Cardinal *num_args)
+{
+    unsigned char *array = *(unsigned char **) to->addr;
+
+    XtFree((char *) array);
 }
 
 /*
@@ -812,7 +942,7 @@ AlignmentArrayDestructor(XtAppContext app,
  */
 /* ARGSUSED */
 Boolean
-CvtStringToPixelTable(Display *dpy,
+XbaeCvtStringToPixelTable(Display *dpy,
 			XrmValuePtr args,
 			Cardinal *num_args,
 			XrmValuePtr from,
@@ -948,7 +1078,7 @@ CvtStringToPixelTable(Display *dpy,
  */
 /* ARGSUSED */
 void
-PixelTableDestructor(XtAppContext app,
+XbaePixelTableDestructor(XtAppContext app,
 			XrmValuePtr to,
 			XtPointer converter_data,
 			XrmValuePtr args,
@@ -960,14 +1090,14 @@ PixelTableDestructor(XtAppContext app,
     if (array)
     {
         for (col = array  ; col ; col++)
-            XtFree((XtPointer)*col);
-        XtFree((XtPointer) array);
+            XtFree((char *)*col);
+        XtFree((char *) array);
     }
 }
 
 /* ARGSUSED */
 Boolean
-CvtStringToGridType(Display *dpy,
+XbaeCvtStringToGridType(Display *dpy,
 		XrmValuePtr args,
 		Cardinal *num_args,
 		XrmValuePtr from,
@@ -998,26 +1128,26 @@ CvtStringToGridType(Display *dpy,
     while (isspace(*start))
 	start++;
 
-    if (StringsAreEqual(start, "grid_none", 9))
+    if (XbaeStringsAreEqual(start, "grid_none", 9))
 	grid_type = XmGRID_NONE;
-    else if (StringsAreEqual(start, "grid_cell_line", 14))
+    else if (XbaeStringsAreEqual(start, "grid_cell_line", 14))
 	grid_type = XmGRID_CELL_LINE;
-    else if (StringsAreEqual(start, "grid_cell_shadow", 16))
+    else if (XbaeStringsAreEqual(start, "grid_cell_shadow", 16))
 	grid_type = XmGRID_CELL_SHADOW;
-    else if (StringsAreEqual(start, "grid_row_line", 13))
+    else if (XbaeStringsAreEqual(start, "grid_row_line", 13))
 	grid_type = XmGRID_ROW_LINE;
-    else if (StringsAreEqual(start, "grid_row_shadow", 15))
+    else if (XbaeStringsAreEqual(start, "grid_row_shadow", 15))
 	grid_type = XmGRID_ROW_SHADOW;
-    else if (StringsAreEqual(start, "grid_column_line", 16))
+    else if (XbaeStringsAreEqual(start, "grid_column_line", 16))
 	grid_type = XmGRID_COLUMN_LINE;
-    else if (StringsAreEqual(start, "grid_column_shadow", 15))
+    else if (XbaeStringsAreEqual(start, "grid_column_shadow", 15))
 	grid_type = XmGRID_COLUMN_SHADOW;
     /* Deprecated types. To be removed in next version. */
-    else if (StringsAreEqual(start, "grid_line", 9))
+    else if (XbaeStringsAreEqual(start, "grid_line", 9))
 	grid_type = XmGRID_LINE;
-    else if (StringsAreEqual(start, "grid_shadow_in", 14))
+    else if (XbaeStringsAreEqual(start, "grid_shadow_in", 14))
 	grid_type = XmGRID_SHADOW_IN;
-    else if (StringsAreEqual(start, "grid_shadow_out", 15))
+    else if (XbaeStringsAreEqual(start, "grid_shadow_out", 15))
 	grid_type = XmGRID_SHADOW_OUT;
     else {
 	XtDisplayStringConversionWarning(dpy, from->addr, XmRGridType);
@@ -1050,9 +1180,9 @@ Boolean
 #ifdef __VMS
 /* According to Barry Stone VMS only allows function names with a
    maximum length of 31 characters */
-CvtStringToMatrixScrollBarDisp
+XbaeCvtStringToMatrixScrollBarD
 #else
-CvtStringToMatrixScrollBarDisplayPolicy
+XbaeCvtStringToMatrixScrollBarDisplayPolicy
 #endif
 	(Display *dpy,
 	XrmValuePtr args,
@@ -1085,11 +1215,11 @@ CvtStringToMatrixScrollBarDisplayPolicy
     while (isspace(*start))
 	start++;
 
-    if (StringsAreEqual(start, "display_none", 12))
+    if (XbaeStringsAreEqual(start, "display_none", 12))
 	display_policy = XmDISPLAY_NONE;
-    else if (StringsAreEqual(start, "display_as_needed", 17))
+    else if (XbaeStringsAreEqual(start, "display_as_needed", 17))
 	display_policy = XmDISPLAY_AS_NEEDED;
-    else if (StringsAreEqual(start, "display_static", 14))
+    else if (XbaeStringsAreEqual(start, "display_static", 14))
 	display_policy = XmDISPLAY_STATIC;
     else {
 	XtDisplayStringConversionWarning(dpy, from->addr,
