@@ -1,6 +1,8 @@
 /*
  * Copyright(c) 1992 Bell Communications Research, Inc. (Bellcore)
  * Copyright(c) 1995-99 Andrew Lister
+ * Copyright  1999, 2000, 2001, 2002, 2003 by the LessTif Developers.
+ *
  *                        All rights reserved
  * Permission to use, copy, modify and distribute this material for
  * any purpose and without fee is hereby granted, provided that the
@@ -22,7 +24,7 @@
  *
  * MatrixWidget Author: Andrew Wason, Bellcore, aw@bae.bellcore.com
  *
- * $Id: Create.c,v 1.9 1999/07/24 02:03:48 lister Exp $
+ * $Id: Create.c,v 1.44 2004/03/04 21:52:27 dannybackx Exp $
  */
 
 /*
@@ -30,62 +32,60 @@
  */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include <XbaeConfig.h>
 #endif
 
+#include <stdio.h>
+#include <stdlib.h>
+
 #include <Xm/Xm.h>
+#include <Xm/ScrollBar.h>
+
 #include <Xbae/MatrixP.h>
 #include <Xbae/Macros.h>
 #include <Xbae/Utils.h>
 #include <Xbae/Actions.h>
-#include <Xm/ScrollBar.h>
 #include <Xbae/Create.h>
 
-static Pixmap createInsensitivePixmap P((XbaeMatrixWidget mw));
+#include <XbaeDebug.h>
 
-#ifndef MAXSCREENS
-#define MAXSCREENS 1
+#if (XmVERSION < 2)
+#define	XmUNSPECIFIED_PIXEL	((Pixel) (~0))
 #endif
 
+static Pixmap createInsensitivePixmap(XbaeMatrixWidget mw);
+
 void
-xbaeCopyBackground(widget, offset, value)
-Widget widget;
-int offset;
-XrmValue *value;
+xbaeCopyBackground(Widget widget, int offset, XrmValue *value)
 {
     value->addr = (XtPointer)&(widget->core.background_pixel);
 }
 
 
 void
-xbaeCopyForeground(widget, offset, value)
-Widget widget;
-int offset;
-XrmValue *value;
+xbaeCopyForeground(Widget widget, int offset, XrmValue *value)
 {
     value->addr = (XtPointer)&(((XmManagerWidget)widget)->manager.foreground);
 }
 
 void
-xbaeCopyDoubleClick(widget, offset, value)
-Widget widget;
-int offset;
-XrmValue *value;
+xbaeCopyDoubleClick(Widget widget, int offset, XrmValue *value)
 {
     static int interval;
-  
+
     interval = XtGetMultiClickTime(XtDisplay(widget));
     value->addr = (XtPointer)&interval;
 }
 
 void
-xbaeCopyCells(mw)
-XbaeMatrixWidget mw;
+xbaeCopyCells(XbaeMatrixWidget mw)
 {
     String **copy = NULL;
     int i, j;
     Boolean empty_row;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows && mw->matrix.columns)
     {
 	/*
@@ -137,16 +137,65 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.cells = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
-#if CELL_WIDGETS
+#ifdef	XBAE_PER_CELL
 void
-xbaeCopyCellWidgets(mw)
-XbaeMatrixWidget mw;
+xbaeFillPerCell(XbaeMatrixWidget mw, struct _XbaeMatrixPerCellRec *p)
+{
+	p->shadow_type = mw->matrix.cell_shadow_type;
+	p->highlighted = HighlightNone;
+	p->selected = False;
+	p->underlined = False;
+	p->user_data = NULL;
+	p->background = XmUNSPECIFIED_PIXEL;
+	p->color = XmUNSPECIFIED_PIXEL;
+	p->widget = NULL;
+	p->font_list = NULL;
+}
+
+void
+xbaeCreatePerCell(XbaeMatrixWidget mw)
+{
+    struct _XbaeMatrixPerCellRec **copy = NULL;
+    int i, j;
+
+    xbaeObjectLock((Widget)mw);
+
+    /*
+     * Malloc an array of row pointers
+     */
+    if (mw->matrix.rows && mw->matrix.columns)
+    {
+	copy = (struct _XbaeMatrixPerCellRec **) XtCalloc((Cardinal)mw->matrix.rows,
+				    sizeof(struct _XbaeMatrixPerCellRec *));
+
+	for (i = 0; i < mw->matrix.rows; i++)
+	{
+		copy[i] = (struct _XbaeMatrixPerCellRec *)
+				XtCalloc((Cardinal)mw->matrix.columns,
+				sizeof(struct _XbaeMatrixPerCellRec));
+		for (j = 0; j < mw->matrix.columns; j++)
+			xbaeFillPerCell(mw, &copy[i][j]);
+
+	}
+    }
+    mw->matrix.per_cell = copy;
+    xbaeObjectUnlock((Widget)mw);
+}
+#endif
+
+#ifndef	XBAE_PER_CELL
+void
+xbaeCopyCellWidgets(XbaeMatrixWidget mw)
 {
     Widget **copy = NULL;
     int i, j;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * Malloc an array of row pointers
      */
@@ -166,16 +215,19 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.cell_widgets = copy;
+    xbaeObjectUnlock((Widget)mw);
 }
 #endif
 
+#ifndef	XBAE_PER_CELL
 void
-xbaeCopyCellShadowTypes(mw)
-XbaeMatrixWidget mw;
+xbaeCopyCellShadowTypes(XbaeMatrixWidget mw)
 {
     unsigned char **copy = NULL;
     int i, j;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows && mw->matrix.columns)
     {
 	copy = (unsigned char **) XtMalloc(mw->matrix.rows *
@@ -203,15 +255,95 @@ XbaeMatrixWidget mw;
 	    }
     }
     mw->matrix.cell_shadow_types = copy;
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyRowShadowTypes(mw)
-XbaeMatrixWidget mw;
+xbaeFreeCellShadowTypes(XbaeMatrixWidget mw)
+{
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.cell_shadow_types)
+    {
+	int i;
+
+	/*
+	 * Free each row of unsigned char pointers
+	 */
+	for (i = 0; i < mw->matrix.rows; i++)
+	    XtFree((char *) mw->matrix.cell_shadow_types[i]);
+
+	/*
+	 * Free the array of row pointers
+	 */
+	XtFree((char *) mw->matrix.cell_shadow_types);
+    }
+    mw->matrix.cell_shadow_types = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+void
+xbaeCopyCellUserData(XbaeMatrixWidget mw)
+{
+    XtPointer **copy = NULL;
+    int i, j;
+
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.rows && mw->matrix.columns)
+    {
+	copy = (XtPointer **) XtMalloc(mw->matrix.rows * sizeof(XtPointer*));
+
+	for (i = 0; i < mw->matrix.rows; i++)
+	    copy[i] = (XtPointer*) XtMalloc(mw->matrix.columns *
+					    sizeof(XtPointer));
+
+	for (i = 0; i < mw->matrix.rows; i++)
+	    for (j = 0; j < mw->matrix.columns; j++)
+		copy[i][j] = mw->matrix.cell_user_data[i][j];
+    }
+    mw->matrix.cell_user_data = copy;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+void
+xbaeFreeCellUserData(XbaeMatrixWidget mw)
+{
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.cell_user_data)
+    {
+	int i;
+
+	/*
+	 * Free each row of Booleans
+	 */
+	for (i = 0; i < mw->matrix.rows; i++)
+	    XtFree((char *) mw->matrix.cell_user_data[i]);
+
+	/*
+	 * Free the array of row pointers
+	 */
+	XtFree((char *) mw->matrix.cell_user_data);
+    }
+    mw->matrix.cell_user_data = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+
+#endif
+
+void
+xbaeCopyRowShadowTypes(XbaeMatrixWidget mw)
 {
     unsigned char *copy = NULL;
     int i;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows)
     {
 	copy = (unsigned char *) XtMalloc(mw->matrix.rows *
@@ -231,15 +363,17 @@ XbaeMatrixWidget mw;
 		copy[i] = mw->matrix.row_shadow_types[i];
     }
     mw->matrix.row_shadow_types = copy;
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnShadowTypes(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColumnShadowTypes(XbaeMatrixWidget mw)
 {
     unsigned char *copy = NULL;
     int i;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.columns)
     {
 	copy = (unsigned char *) XtMalloc(mw->matrix.columns *
@@ -259,38 +393,17 @@ XbaeMatrixWidget mw;
 		copy[i] = mw->matrix.column_shadow_types[i];
     }
     mw->matrix.column_shadow_types = copy;
-}
-
-
-void
-xbaeCopyCellUserData(mw)
-XbaeMatrixWidget mw;
-{
-    XtPointer **copy = NULL;
-    int i, j;
-
-    if (mw->matrix.rows && mw->matrix.columns)
-    {
-	copy = (XtPointer **) XtMalloc(mw->matrix.rows * sizeof(XtPointer*));
-
-	for (i = 0; i < mw->matrix.rows; i++)
-	    copy[i] = (XtPointer*) XtMalloc(mw->matrix.columns *
-					    sizeof(XtPointer));
-
-	for (i = 0; i < mw->matrix.rows; i++)
-	    for (j = 0; j < mw->matrix.columns; j++)
-		copy[i][j] = mw->matrix.cell_user_data[i][j];
-    }
-    mw->matrix.cell_user_data = copy;
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyRowUserData(mw)
-XbaeMatrixWidget mw;
+xbaeCopyRowUserData(XbaeMatrixWidget mw)
 {
     XtPointer *copy = NULL;
     int i;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows)
     {
 	copy = (XtPointer *) XtMalloc(mw->matrix.rows * sizeof(XtPointer));
@@ -299,15 +412,18 @@ XbaeMatrixWidget mw;
 	    copy[i] = mw->matrix.row_user_data[i];
     }
     mw->matrix.row_user_data = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnUserData(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColumnUserData(XbaeMatrixWidget mw)
 {
     XtPointer *copy = NULL;
     int i;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.columns)
     {
 	copy = (XtPointer *) XtMalloc(mw->matrix.columns * sizeof(XtPointer));
@@ -316,16 +432,19 @@ XbaeMatrixWidget mw;
 	    copy[i] = mw->matrix.column_user_data[i];
     }
     mw->matrix.column_user_data = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyRowLabels(mw)
-XbaeMatrixWidget mw;
+xbaeCopyRowLabels(XbaeMatrixWidget mw)
 {
     String *copy = NULL;
     int i;
     Boolean empty_label;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows)
     {
 	copy = (String *) XtMalloc(mw->matrix.rows * sizeof(String));
@@ -345,17 +464,21 @@ XbaeMatrixWidget mw;
 		copy[i] = XtNewString(mw->matrix.row_labels[i]);
     }
     mw->matrix.row_labels = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnLabels(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColumnLabels(XbaeMatrixWidget mw)
 {
-    String *copy = NULL;
-    int i;
-    Boolean empty_column;
-    
-    if (mw->matrix.columns)
+	String *copy = NULL;
+	XmString	*xmcopy;
+	int i;
+	Boolean empty_column;
+
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.columns && mw->matrix.column_labels)
     {
 	copy = (String *) XtMalloc(mw->matrix.columns * sizeof(String));
 
@@ -395,16 +518,27 @@ XbaeMatrixWidget mw;
 		    mw->matrix.column_label_lines[i].lines;
     }
     mw->matrix.column_labels = copy;
+
+/* Straight through for xmcolumn_label */
+	if (mw->matrix.columns && mw->matrix.xmcolumn_labels) {
+		xmcopy = (XmString *)XtMalloc(mw->matrix.columns * sizeof(XmString));
+		for (i=0; i<mw->matrix.columns; i++) {
+			xmcopy[i] = XmStringCopy(mw->matrix.xmcolumn_labels[i]);
+		}
+	}
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnWidths(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColumnWidths(XbaeMatrixWidget mw)
 {
     short *copy = NULL;
     int i;
     Boolean bad = False;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.columns)
     {
 	copy = (short *) XtMalloc(mw->matrix.columns * sizeof(short));
@@ -428,16 +562,57 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.column_widths = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnMaxLengths(mw)
-XbaeMatrixWidget mw;
+xbaeCopyRowHeights(XbaeMatrixWidget mw)
+{
+    short *copy = NULL;
+    int i;
+    Boolean bad = False;
+
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.rows)
+    {
+	copy = (short *) XtMalloc(mw->matrix.rows * sizeof(short));
+
+	for (i = 0; i < mw->matrix.rows; i++)
+	{
+	    if (!bad && mw->matrix.row_heights[i] == BAD_HEIGHT)
+	    {
+		bad = True;
+		XtAppWarningMsg(
+		    XtWidgetToApplicationContext((Widget) mw),
+		    "copyRowHeights", "tooShort", "XbaeMatrix",
+		    "XbaeMatrix: Row heights array is too short",
+		    NULL, 0);
+		copy[i] = 1;
+	    }
+	    else if (bad)
+		copy[i] = 1;
+	    else
+		copy[i] = mw->matrix.row_heights[i];
+	}
+    }
+    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)mw, "xbaeCopyRowHeights %p (old %p)\n",
+			 copy, mw->matrix.row_heights));
+    mw->matrix.row_heights = copy;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+void
+xbaeCopyColumnMaxLengths(XbaeMatrixWidget mw)
 {
     int *copy = NULL;
     int i;
     Boolean bad = False;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.columns)
     {
 	copy = (int *) XtMalloc(mw->matrix.columns * sizeof(int));
@@ -461,16 +636,19 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.column_max_lengths = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnAlignments(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColumnAlignments(XbaeMatrixWidget mw)
 {
     unsigned char *copy = NULL;
     int i;
     Boolean bad = False;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.columns)
     {
 	copy = (unsigned char *) XtMalloc(mw->matrix.columns *
@@ -495,15 +673,18 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.column_alignments = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnButtonLabels(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColumnButtonLabels(XbaeMatrixWidget mw)
 {
     Boolean *copy = NULL;
     int i;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.columns)
     {
 	copy = (Boolean *) XtMalloc(mw->matrix.columns *
@@ -515,15 +696,63 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.column_button_labels = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyRowButtonLabels(mw)
-XbaeMatrixWidget mw;
+xbaeCopyShowColumnArrows(XbaeMatrixWidget mw)
+{
+	Boolean *copy = NULL;
+	int i;
+
+	xbaeObjectLock((Widget)mw);
+
+	if (mw->matrix.columns) {
+		copy = (Boolean *) XtMalloc(mw->matrix.columns * sizeof(Boolean));
+
+		for (i = 0; i < mw->matrix.columns; i++) {
+			copy[i] = mw->matrix.show_column_arrows[i];
+		}
+		xbaeFreeShowColumnArrows(mw);
+	}
+	mw->matrix.show_column_arrows = copy;
+
+	xbaeObjectUnlock((Widget)mw);
+}
+
+
+void
+xbaeCopyColumnFontBold(XbaeMatrixWidget mw)
 {
     Boolean *copy = NULL;
     int i;
 
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.columns)
+    {
+      copy = (Boolean *) XtMalloc(mw->matrix.columns *
+                                  sizeof(Boolean));
+
+      for (i = 0; i < mw->matrix.columns; i++)
+      {
+          copy[i] = mw->matrix.column_font_bold[i];
+      }
+    }
+    mw->matrix.column_font_bold = copy;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+void
+xbaeCopyRowButtonLabels(XbaeMatrixWidget mw)
+{
+    Boolean *copy = NULL;
+    int i;
+
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows)
     {
 	copy = (Boolean *) XtMalloc(mw->matrix.rows *
@@ -535,16 +764,19 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.row_button_labels = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCopyColumnLabelAlignments(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColumnLabelAlignments(XbaeMatrixWidget mw)
 {
     unsigned char *copy = NULL;
     int i;
     Boolean bad = False;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.columns)
     {
 	copy = (unsigned char *) XtMalloc(mw->matrix.columns *
@@ -571,17 +803,20 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.column_label_alignments = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
+#ifndef	XBAE_PER_CELL
 void
-xbaeCopyColors(mw)
-XbaeMatrixWidget mw;
+xbaeCopyColors(XbaeMatrixWidget mw)
 {
     Pixel **copy = NULL;
     int i, j;
     Boolean badrow = False;
     Boolean badcol;
 
+    xbaeObjectLock((Widget)mw);
 
     if (mw->matrix.rows && mw->matrix.columns)
     {
@@ -634,17 +869,74 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.colors = copy;
-}
 
+    xbaeObjectUnlock((Widget)mw);
+}
+#endif
+
+#ifdef	XBAE_PER_CELL
+/*
+ * Copy the per_cell array, which holds all the per-cell information
+ * except the cell string itself.
+ *
+ * This means that the per_cell structure gets allocated for every cell that
+ * has one of these (mostly visual) resources defined.
+ *
+ * Sharing all this for all the per-cell fields means cutting all the overhead.
+ * Also all of the double code gets reduced.
+ */
 void
-xbaeCopyBackgrounds(mw)
-XbaeMatrixWidget mw;
+xbaeCopyPerCell(XbaeMatrixWidget mw)
+{
+    struct _XbaeMatrixPerCellRec	**copy = NULL;
+    int i, j;
+
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.rows && mw->matrix.columns)
+    {
+	/*
+	 * Malloc an array of row pointers
+	 */
+	copy = (struct _XbaeMatrixPerCellRec **) XtCalloc(mw->matrix.rows, sizeof(struct _XbaeMatrixPerCellRec *));
+
+	/*
+	 * Malloc an array of _XbaeMatrixPerCellRec for each row pointer
+	 */
+	for (i = 0; i < mw->matrix.rows; i++)
+	    copy[i] = (struct _XbaeMatrixPerCellRec *) XtCalloc(mw->matrix.columns,
+					   sizeof(struct _XbaeMatrixPerCellRec));
+
+	/*
+	 * If per_cell is not NULL, copy the table passed in
+	 */
+	if (mw->matrix.per_cell) {
+	    for (i = 0; i < mw->matrix.rows; i++)
+		for (j = 0; j < mw->matrix.columns; j++)
+		    copy[i][j] = mw->matrix.per_cell[i][j];
+	} else {
+	    for (i = 0; i < mw->matrix.rows; i++)
+		for (j = 0; j < mw->matrix.columns; j++)
+			xbaeFillPerCell(mw, &copy[i][j]);
+	}
+    }
+    mw->matrix.per_cell = copy;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+#endif
+
+#ifndef	XBAE_PER_CELL
+void
+xbaeCopyBackgrounds(XbaeMatrixWidget mw)
 {
     Pixel **copy = NULL;
     int i, j;
     Boolean badrow = False;
     Boolean badcol;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows && mw->matrix.columns)
     {
 	/*
@@ -662,20 +954,11 @@ XbaeMatrixWidget mw;
 	{
 	    for (i = 0; i < mw->matrix.rows; i++)
 	    {
-		Boolean alt = (mw->matrix.alt_row_count &&
-			       i >= (int)mw->matrix.fixed_rows) ? (
-				   ((i - (int)mw->matrix.fixed_rows) /
-				    mw->matrix.alt_row_count) % 2) : False;
-
-		/*
-		 * Assign the even and odd row colours appropriately.  These
-		 * will be a copy of the core->background if they have not
-		 * been explicitly set but if they have, we want to
-		 * preserve the colours as they appear now
-		 */
-		for (j = 0; j < mw->matrix.columns; j++)
-		    copy[i][j] = (alt ? mw->matrix.odd_row_background :
-				  mw->matrix.even_row_background);
+			for (j = 0; j < mw->matrix.columns; j++)
+				/**
+				 * xbaeComputeCellColors() will generate the stripe
+				 */
+				copy[i][j] = XmUNSPECIFIED_PIXEL;
 	    }
 	}
 	else for (i = 0; i < mw->matrix.rows; i++)
@@ -710,6 +993,8 @@ XbaeMatrixWidget mw;
 	}
     }
     mw->matrix.cell_background = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 /*
@@ -717,19 +1002,24 @@ XbaeMatrixWidget mw;
  * represent selected cells if it is NULL.
  */
 void
-xbaeCopySelectedCells(mw)
-XbaeMatrixWidget mw;
+xbaeCopySelectedCells(XbaeMatrixWidget mw)
 {
     Boolean **copy = NULL;
     int i, j;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows && mw->matrix.columns)
     {
 	/*
 	 * Malloc an array of row pointers
 	 */
 	mw->matrix.num_selected_cells = 0;
+#if 0
 	copy = (Boolean **) XtMalloc(mw->matrix.rows * sizeof(Boolean *));
+#else
+	copy = (Boolean **) XtCalloc(mw->matrix.rows, sizeof(Boolean *));
+#endif
 
 	/*
 	 * Malloc an array of Booleans for each row pointer
@@ -751,20 +1041,65 @@ XbaeMatrixWidget mw;
 		}
     }
     mw->matrix.selected_cells = copy;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+/*
+ * Copy the underlinedCells resource. Create a 2D array of Booleans to
+ * represent underlined cells if it is NULL.
+ */
+void
+xbaeCopyUnderlinedCells(XbaeMatrixWidget mw)
+{
+    Boolean **copy = NULL;
+    int i, j;
+
+    xbaeObjectLock((Widget)mw);
+
+    if (mw->matrix.rows && mw->matrix.columns)
+    {
+	/*
+	 * Malloc an array of row pointers
+	 */
+#if 0
+	copy = (Boolean **) XtMalloc(mw->matrix.rows * sizeof(Boolean *));
+#else
+	copy = (Boolean **) XtCalloc(mw->matrix.rows, sizeof(Boolean *));
+#endif
+
+	/*
+	 * Malloc an array of Booleans for each row pointer
+	 */
+	for (i = 0; i < mw->matrix.rows; i++)
+	    copy[i] = (Boolean *) XtCalloc(mw->matrix.columns,
+					   sizeof(Boolean));
+
+	/*
+	 * If underlined_cells is not NULL, copy the table passed in
+	 */
+	if (mw->matrix.underlined_cells)
+	    for (i = 0; i < mw->matrix.rows; i++)
+		for (j = 0; j < mw->matrix.columns; j++)
+		    copy[i][j] = mw->matrix.underlined_cells[i][j];
+    }
+    mw->matrix.underlined_cells = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
-#if XmVersion >= 1002
 /*
  * Copy the highlightedCells resource. Create a 2D array of Booleans to
  * represent highlighted cells if it is NULL.
  */
 void
-xbaeCopyHighlightedCells(mw)
-XbaeMatrixWidget mw;
+xbaeCopyHighlightedCells(XbaeMatrixWidget mw)
 {
     unsigned char **copy = NULL;
     int i, j;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows && mw->matrix.columns)
     {
 	/*
@@ -789,18 +1124,20 @@ XbaeMatrixWidget mw;
 		    copy[i][j] = mw->matrix.highlighted_cells[i][j];
     }
     mw->matrix.highlighted_cells = copy;
+
+    xbaeObjectUnlock((Widget)mw);
 }
-#endif
 
 /*
  * Create a matrix of Pixels
  */
 void
-xbaeCreateColors(mw)
-XbaeMatrixWidget mw;
+xbaeCreateColors(XbaeMatrixWidget mw)
 {
     int i;
 
+    xbaeObjectLock((Widget)mw);
+
     if (mw->matrix.rows && mw->matrix.columns)
     {
 	/*
@@ -818,6 +1155,80 @@ XbaeMatrixWidget mw;
     }
     else
 	mw->matrix.colors = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+#endif
+
+/*
+ * A cache for pixmaps which remembers the screen a pixmap belongs to
+ * (see bug #591306) and that permits clearing by screen when that
+ * screen gets closed.
+ */
+static struct pcache {
+	Pixmap	pixmap;
+	Screen	*scr;
+
+} *stipple_cache = NULL;
+static int	ncache = 0;	/* Allocated size of the array. */
+				/* Empty places have NULL screen */
+
+static Pixmap
+PixmapFromCache(Screen *scr)
+{
+	int	i;
+
+	for (i=0; i<ncache; i++)
+		if (scr == stipple_cache[i].scr)
+			return stipple_cache[i].pixmap;
+	return (Pixmap)0;
+}
+
+static void
+AddPixmapToCache(Screen *scr, Pixmap p)
+{
+	int	i, old;
+
+	for (i=0; i<ncache; i++)
+		if (stipple_cache[i].scr == 0) {
+			stipple_cache[i].scr = scr;
+			stipple_cache[i].pixmap = p;
+			return;
+		}
+
+	/* Allocate more */
+	if (ncache) {
+		old = ncache;
+		ncache *= 2;
+		stipple_cache = (struct pcache *)XtRealloc((char *)stipple_cache,
+			ncache * sizeof(struct pcache));
+		for (i=old; i<ncache; i++)
+			stipple_cache[i].scr = NULL;
+		stipple_cache[old].scr = scr;
+		stipple_cache[old].pixmap = p;
+	} else {
+		ncache = 16;	/* Some silly initial value */
+		stipple_cache = (struct pcache *)XtCalloc(ncache,
+			sizeof(struct pcache));
+		stipple_cache[0].scr = scr;
+		stipple_cache[0].pixmap = p;
+	}
+}
+
+/*
+ * Remove the pixmaps with this screen from the cache
+ */
+static void
+RemovePixmapsFromScreen(Screen *scr)
+{
+	int	i;
+	for (i=0; i<ncache; i++)
+		if (stipple_cache[i].scr == scr) {
+			XFreePixmap(DisplayOfScreen(stipple_cache[i].scr),
+				stipple_cache[i].pixmap);
+			stipple_cache[i].pixmap = (Pixmap)0;
+			stipple_cache[i].scr = NULL;
+		}
 }
 
 /*
@@ -825,63 +1236,67 @@ XbaeMatrixWidget mw;
  * XmNsensitive is set to False
  */
 static Pixmap
-createInsensitivePixmap(mw)
-XbaeMatrixWidget mw;
+createInsensitivePixmap(XbaeMatrixWidget mw)
 {
-    static char stippleBits[] = { 0x01, 0x02 };
-    static Pixmap stipple[MAXSCREENS] = {(Pixmap)NULL};
-    Display *dpy = XtDisplay(mw);
-    Screen *scr  = XtScreen (mw);
-    int i;
-    int maxScreens = ScreenCount(dpy);
-    
-    if (maxScreens > MAXSCREENS)
-	maxScreens = MAXSCREENS;
-    
-    if (!stipple[0])
-    {
-	for (i = 0 ; i < maxScreens ; i++)
-	    stipple[i] = XCreatePixmapFromBitmapData(
-		dpy, RootWindow(dpy,i), stippleBits, 2, 2, 0, 1, 1);
-    }
-    for (i = 0; i < maxScreens; i++)
-    {
-	if (ScreenOfDisplay(dpy, i) == scr)
-	    return stipple[i];
-    }
-    return (Pixmap)NULL;
+	static char stippleBits[] = { 0x01, 0x02 };
+	Display *dpy = XtDisplay(mw);
+	Screen *scr  = XtScreen (mw);
+	Pixmap	p;
+
+	xbaeObjectLock((Widget)mw);
+
+	p = PixmapFromCache(XtScreen((Widget)mw));
+	if (p) {
+		xbaeObjectUnlock((Widget)mw);
+		return p;
+	}
+
+	p = XCreatePixmapFromBitmapData(dpy, RootWindowOfScreen(scr),
+		stippleBits, 2, 2, 0, 1, 1);
+	AddPixmapToCache(scr, p);
+	xbaeObjectUnlock((Widget)mw);
+	return p;
 }
-    
+
 void
-xbaeCreateGridLineGC(mw)
-XbaeMatrixWidget mw;
+xbaeCreateGridLineGC(XbaeMatrixWidget mw)
 {
     XGCValues values;
     XtGCMask mask = GCForeground | GCBackground;
 
+    xbaeObjectLock((Widget)mw);
+
     values.foreground = mw->matrix.grid_line_color;
     values.background = mw->manager.foreground;
-    
+
+    /* Free em before recreating */
+    if (mw->matrix.grid_line_gc)
+	XtReleaseGC((Widget)mw, mw->matrix.grid_line_gc);
+    if (mw->matrix.cell_grid_line_gc)
+	XFreeGC(XtDisplay(mw), mw->matrix.cell_grid_line_gc);
     /*
      * GC for drawing grid lines
      */
     mw->matrix.grid_line_gc = XtGetGC((Widget) mw, mask, &values);
-    
+
     /*
      * GC for drawing grid lines with clipping
      */
     mw->matrix.cell_grid_line_gc = XCreateGC(XtDisplay(mw),
 					     GC_PARENT_WINDOW(mw),
 					     mask, &values);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCreateDrawGC(mw)
-XbaeMatrixWidget mw;
+xbaeCreateDrawGC(XbaeMatrixWidget mw)
 {
     XGCValues values;
     unsigned long mask = GCForeground | GCFont | GCStipple;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * GC for drawing cells. We create it instead of using a cached one,
      * since the foreground may change frequently.
@@ -890,18 +1305,29 @@ XbaeMatrixWidget mw;
     values.font = mw->matrix.fid;
     values.stipple = createInsensitivePixmap(mw);
 
+    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)mw,
+	"xbaeCreateDrawGC(dpy %p win %p fg %d font %p stip %p)\n",
+	XtDisplay(mw),
+	GC_PARENT_WINDOW(mw),
+	values.foreground,
+	values.font,
+	values.stipple));
+
     mw->matrix.draw_gc = XCreateGC(XtDisplay(mw),
 				   GC_PARENT_WINDOW(mw),
 				   mask, &values);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCreatePixmapGC(mw)
-XbaeMatrixWidget mw;
+xbaeCreatePixmapGC(XbaeMatrixWidget mw)
 {
     XGCValues values;
     unsigned long mask = GCForeground | GCGraphicsExposures | GCStipple;
 
+    xbaeObjectLock((Widget)mw);
+
     values.foreground = mw->manager.foreground;
     values.graphics_exposures = False;
     values.stipple = createInsensitivePixmap(mw);
@@ -909,15 +1335,18 @@ XbaeMatrixWidget mw;
     mw->matrix.pixmap_gc = XCreateGC(XtDisplay(mw),
 				     GC_PARENT_WINDOW(mw),
 				     mask, &values);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCreateLabelGC(mw)
-XbaeMatrixWidget mw;
+xbaeCreateLabelGC(XbaeMatrixWidget mw)
 {
     XGCValues values;
     unsigned long mask = GCForeground | GCFont | GCStipple;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * GC for drawing labels
      */
@@ -927,15 +1356,18 @@ XbaeMatrixWidget mw;
     mw->matrix.label_gc = XCreateGC(XtDisplay(mw),
 				    GC_PARENT_WINDOW(mw),
 				    mask, &values);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCreateLabelClipGC(mw)
-XbaeMatrixWidget mw;
+xbaeCreateLabelClipGC(XbaeMatrixWidget mw)
 {
     XGCValues values;
     unsigned long mask = GCForeground | GCFont | GCStipple;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * GC for drawing labels with clipping.
      */
@@ -945,15 +1377,18 @@ XbaeMatrixWidget mw;
     mw->matrix.label_clip_gc = XCreateGC(XtDisplay(mw),
 					 GC_PARENT_WINDOW(mw),
 					 mask, &values);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCreateTopShadowClipGC(mw)
-XbaeMatrixWidget mw;
+xbaeCreateTopShadowClipGC(XbaeMatrixWidget mw)
 {
     XGCValues values;
     XtGCMask mask = GCForeground | GCBackground;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * GC for drawing top shadow inside cells with clipping.
      */
@@ -973,15 +1408,18 @@ XbaeMatrixWidget mw;
     values.function = GXxor;
     mw->matrix.resize_top_shadow_gc = XtGetGC(
 	(Widget) mw, mask, &values);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeCreateBottomShadowClipGC(mw)
-XbaeMatrixWidget mw;
+xbaeCreateBottomShadowClipGC(XbaeMatrixWidget mw)
 {
     XGCValues values;
     XtGCMask mask = GCForeground | GCBackground;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * GC for drawing bottom shadow inside cells with clipping.
      */
@@ -1001,11 +1439,12 @@ XbaeMatrixWidget mw;
     values.function = GXxor;
     mw->matrix.resize_bottom_shadow_gc = XtGetGC(
 	(Widget) mw, mask, &values);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeNewFont(mw)
-XbaeMatrixWidget mw;
+xbaeNewFont(XbaeMatrixWidget mw, int frominit)
 {
     XmFontContext context;
     XFontStruct *font;
@@ -1015,10 +1454,13 @@ XbaeMatrixWidget mw;
     XFontStruct **fonts;
     char **font_names;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * Make a private copy of the FontList
      */
-    mw->matrix.font_list = XmFontListCopy(mw->matrix.font_list);
+    if (! frominit)
+	mw->matrix.font_list = XmFontListCopy(mw->matrix.font_list);
 
     /*
      * Get XmFontListEntry from FontList
@@ -1065,11 +1507,12 @@ XbaeMatrixWidget mw;
     }
 
     XmFontListFreeFontContext(context);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeNewLabelFont(mw)
-XbaeMatrixWidget mw;
+xbaeNewLabelFont(XbaeMatrixWidget mw)
 {
     XmFontContext context;
     XFontStruct *font;
@@ -1079,6 +1522,8 @@ XbaeMatrixWidget mw;
     XFontStruct **fonts;
     char **font_names;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * Make a private copy of the FontList
      */
@@ -1129,146 +1574,148 @@ XbaeMatrixWidget mw;
     }
 
     XmFontListFreeFontContext(context);
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeFreeCells(mw)
-XbaeMatrixWidget mw;
+xbaeFreeCells(XbaeMatrixWidget mw)
 {
     int i, j;
 
     if (!mw->matrix.cells)
 	return;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * Free each cell in a row, then free the row and go to the next one
      */
     for (i = 0; i < mw->matrix.rows; i++)
     {
 	for (j = 0; j < mw->matrix.columns; j++)
-	    XtFree((XtPointer) mw->matrix.cells[i][j]);
-	XtFree((XtPointer) mw->matrix.cells[i]);
+	    XtFree((char *) mw->matrix.cells[i][j]);
+	XtFree((char *) mw->matrix.cells[i]);
     }
 
     /*
      * Free the array of row pointers
      */
-    XtFree((XtPointer) mw->matrix.cells);
+    XtFree((char *) mw->matrix.cells);
     mw->matrix.cells = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
-#if CELL_WIDGETS
+#ifdef	XBAE_PER_CELL
 void
-xbaeFreeCellWidgets(mw)
-XbaeMatrixWidget mw;
+xbaeFreePerCell(XbaeMatrixWidget mw)
+{
+    int i;
+
+    /*
+     * Free each row of XtPointer pointers
+     */
+    if (!mw->matrix.per_cell)
+	return;
+
+    xbaeObjectLock((Widget)mw);
+
+    for (i = 0; i < mw->matrix.rows; i++)
+	XtFree((char *) mw->matrix.per_cell[i]);
+
+    /*
+     * Free the array of row pointers
+     */
+    XtFree((char *) mw->matrix.per_cell);
+    mw->matrix.per_cell = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+#else
+void
+xbaeFreeCellWidgets(XbaeMatrixWidget mw)
 {
     int i, j;
 
     if (!mw->matrix.cell_widgets)
 	return;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
-     * Free each cell in a row, then free the row and go to the next one
+     * Destroy each cell widget (don't XtFree ! as Linas suggests),
+     * then free the row and go to the next one.
      */
     for (i = 0; i < mw->matrix.rows; i++)
     {
 	for (j = 0; j < mw->matrix.columns; j++)
-	    XtFree((XtPointer) mw->matrix.cell_widgets[i][j]);
-	XtFree((XtPointer) mw->matrix.cell_widgets[i]);
+	    if (mw->matrix.cell_widgets[i][j])
+		XtDestroyWidget(mw->matrix.cell_widgets[i][j]);
+	XtFree((char *) mw->matrix.cell_widgets[i]);
     }
 
     /*
      * Free the array of row pointers
      */
-    XtFree((XtPointer) mw->matrix.cell_widgets);
+    XtFree((char *) mw->matrix.cell_widgets);
     mw->matrix.cell_widgets = NULL;
-}
-#endif
 
-void
-xbaeFreeRowLabels(mw)
-XbaeMatrixWidget mw;
-{
-    int i;
-
-    if (!mw->matrix.row_labels)
-	return;
-
-    for (i = 0; i < mw->matrix.rows; i++)
-	XtFree((XtPointer) mw->matrix.row_labels[i]);
-
-    XtFree((XtPointer) mw->matrix.row_labels);
-    mw->matrix.row_labels = NULL;
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeFreeColumnLabels(mw)
-XbaeMatrixWidget mw;
-{
-    int i;
-
-    if (!mw->matrix.column_labels)
-	return;
-
-    for (i = 0; i < mw->matrix.columns; i++)
-    {
-	XtFree((XtPointer) mw->matrix.column_labels[i]);
-	XtFree((XtPointer) mw->matrix.column_label_lines[i].lengths);
-    }
-
-    XtFree((XtPointer) mw->matrix.column_label_lines);
-    XtFree((XtPointer) mw->matrix.column_labels);
-    mw->matrix.column_labels = NULL;
-}
-
-
-void
-xbaeFreeColors(mw)
-XbaeMatrixWidget mw;
+xbaeFreeColors(XbaeMatrixWidget mw)
 {
     int i;
 
     if (!mw->matrix.colors)
 	return;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * Free each row of Pixels
      */
     for (i = 0; i < mw->matrix.rows; i++)
-	XtFree((XtPointer) mw->matrix.colors[i]);
+	XtFree((char *) mw->matrix.colors[i]);
 
     /*
      * Free the array of row pointers
      */
-    XtFree((XtPointer) mw->matrix.colors);
+    XtFree((char *) mw->matrix.colors);
     mw->matrix.colors = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeFreeBackgrounds(mw)
-XbaeMatrixWidget mw;
+xbaeFreeBackgrounds(XbaeMatrixWidget mw)
 {
     int i;
 
     if (!mw->matrix.cell_background)
 	return;
 
+    xbaeObjectLock((Widget)mw);
+
     /*
      * Free each row of Pixels
      */
     for (i = 0; i < mw->matrix.rows; i++)
-	XtFree((XtPointer) mw->matrix.cell_background[i]);
+	XtFree((char *) mw->matrix.cell_background[i]);
 
     /*
      * Free the array of row pointers
      */
-    XtFree((XtPointer) mw->matrix.cell_background);
+    XtFree((char *) mw->matrix.cell_background);
     mw->matrix.cell_background = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
 void
-xbaeFreeSelectedCells(mw)
-XbaeMatrixWidget mw;
+xbaeFreeSelectedCells(XbaeMatrixWidget mw)
 {
     int i;
 
@@ -1278,82 +1725,147 @@ XbaeMatrixWidget mw;
     if (!mw->matrix.selected_cells)
 	return;
 
+    xbaeObjectLock((Widget)mw);
+
     for (i = 0; i < mw->matrix.rows; i++)
-	XtFree((XtPointer) mw->matrix.selected_cells[i]);
+	XtFree((char *) mw->matrix.selected_cells[i]);
 
     /*
      * Free the array of row pointers
      */
-    XtFree((XtPointer) mw->matrix.selected_cells);
+    XtFree((char *) mw->matrix.selected_cells);
     mw->matrix.selected_cells = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+#endif
+
+void
+xbaeFreeRowLabels(XbaeMatrixWidget mw)
+{
+    int i;
+
+    if (!mw->matrix.row_labels)
+	return;
+
+    xbaeObjectLock((Widget)mw);
+
+    for (i = 0; i < mw->matrix.rows; i++)
+	XtFree((char *) mw->matrix.row_labels[i]);
+
+    XtFree((char *) mw->matrix.row_labels);
+    mw->matrix.row_labels = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
 }
 
-#if XmVersion >= 1002
+/* FIX ME COLUMN_LABEL */
 void
-xbaeFreeHighlightedCells(mw)
-XbaeMatrixWidget mw;
+xbaeFreeColumnLabels(XbaeMatrixWidget mw)
+{
+    int i;
+
+    if (!mw->matrix.column_labels)
+	return;
+
+    xbaeObjectLock((Widget)mw);
+
+    for (i = 0; i < mw->matrix.columns; i++)
+    {
+	XtFree((char *) mw->matrix.column_labels[i]);
+	XtFree((char *) mw->matrix.column_label_lines[i].lengths);
+    }
+
+    XtFree((char *) mw->matrix.column_label_lines);
+    XtFree((char *) mw->matrix.column_labels);
+    mw->matrix.column_labels = NULL;
+    mw->matrix.column_label_lines = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+#ifndef	XBAE_PER_CELL
+void
+xbaeFreeUnderlinedCells(XbaeMatrixWidget mw)
+{
+    int i;
+
+    /*
+     * Free each row of XtPointer pointers
+     */
+    if (!mw->matrix.underlined_cells)
+	return;
+
+    xbaeObjectLock((Widget)mw);
+
+    for (i = 0; i < mw->matrix.rows; i++)
+	XtFree((char *) mw->matrix.underlined_cells[i]);
+
+    /*
+     * Free the array of row pointers
+     */
+    XtFree((char *) mw->matrix.underlined_cells);
+    mw->matrix.underlined_cells = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+
+void
+xbaeFreeHighlightedCells(XbaeMatrixWidget mw)
 {
     int i;
 
     if (!mw->matrix.highlighted_cells)
 	return;
-    
+
+    xbaeObjectLock((Widget)mw);
+
     /*
      * Free each row of XtPointer pointers
      */
     for (i = 0; i < mw->matrix.rows; i++)
-	XtFree((XtPointer) mw->matrix.highlighted_cells[i]);
+	XtFree((char *) mw->matrix.highlighted_cells[i]);
 
     /*
      * Free the array of row pointers
      */
-    XtFree((XtPointer) mw->matrix.highlighted_cells);
+    XtFree((char *) mw->matrix.highlighted_cells);
     mw->matrix.highlighted_cells = NULL;
+
+    xbaeObjectUnlock((Widget)mw);
+}
+#endif
+
+/*
+ * Make sure to know when our display connection dies.
+ */
+#if XtSpecificationRelease >= 6
+static void
+DisplayDied(Widget w, XtPointer client, XtPointer call)
+{
+	XtDestroyHookDataRec	*p = (XtDestroyHookDataRec *)call;
+
+	if (p == NULL || p->type != XtHdestroy)
+		return;
+
+	if (XtIsSubclass(p->widget, xmPrimitiveWidgetClass))
+		RemovePixmapsFromScreen(XtScreen(p->widget));
 }
 #endif
 
 void
-xbaeFreeCellUserData(mw)
-XbaeMatrixWidget mw;
+xbaeRegisterDisplay(Widget w)
 {
-    if (mw->matrix.cell_user_data)
-    {
-	int i;
+#if XtSpecificationRelease >= 6
+	Display	*d = XtDisplay(w);
 
-	/*
-	 * Free each row of Booleans
-	 */
-	for (i = 0; i < mw->matrix.rows; i++)
-	    XtFree((XtPointer) mw->matrix.cell_user_data[i]);
+/*
+* Hummingbird Exceed Motif libraries have a bug
+* XtSpecificationRelease is 6, but XtHooksOfDisplay is missing :(
+*/
+#ifndef EXCEED_MOTIF
+	XtAddCallback(XtHooksOfDisplay(d), XtNdestroyHook, DisplayDied, NULL);
+#endif
 
-	/*
-	 * Free the array of row pointers
-	 */
-	XtFree((XtPointer) mw->matrix.cell_user_data);
-    }
-    mw->matrix.cell_user_data = NULL;
+#endif
 }
-
-
-void
-xbaeFreeCellShadowTypes(mw)
-XbaeMatrixWidget mw;
-{
-    if (mw->matrix.cell_shadow_types)
-    {
-	int i;
-
-	/*
-	 * Free each row of unsigned char pointers
-	 */
-	for (i = 0; i < mw->matrix.rows; i++)
-	    XtFree((XtPointer) mw->matrix.cell_shadow_types[i]);
-
-	/*
-	 * Free the array of row pointers
-	 */
-	XtFree((XtPointer) mw->matrix.cell_shadow_types);
-    }
-    mw->matrix.cell_shadow_types = NULL;
-}
-
