@@ -1,5 +1,6 @@
 /*
- * Copyright(c) 1999 Andrew Lister
+ * Copyright (c) 1999 Andrew Lister
+ * Copyright  1999, 2000, 2001, 2002, 2003 by the LessTif Developers.
  *
  *                        All rights reserved
  * Permission to use, copy, modify and distribute this material for
@@ -21,35 +22,24 @@
  *
  * Author: Andrew Lister
  *
- * $Id: Input.c,v 1.4 2000/02/29 19:29:23 danny Exp $
+ * $Id: Input.c,v 1.25 2003/06/08 12:20:15 dannybackx Exp $
  */
 
-/*#define DEBUG*/
-
 #ifdef HAVE_CONFIG_H
 #include <XbaeConfig.h>
 #endif
 
-/* For memmove/bcopy */
+#include <stdio.h>
+#include <ctype.h>
 #include <string.h>
 
-#include <ctype.h>
-#include <Xbae/InputP.h>
-#if XmVersion <= 1001
-#include <Xm/XmP.h>
-#else
 #include <Xm/DrawP.h>
-#endif
+#include <Xm/ScrollBar.h>
 
-#if !defined(HAVE_MEMMOVE) && !defined(XBAE_NEED_BCOPY)
-#define XBAE_NEED_BCOPY
-#endif
+#include <Xbae/InputP.h>
+
+#include <XbaeDebug.h>
 
-#if !defined(XBAE_NEED_BCOPY) || defined(SVR4) || defined(VMS)
-#define BCOPY(src, dest, n)     memmove((void *)(dest), (void *)(src), (n))
-#else
-#define BCOPY(src, dest, n)     bcopy((void *)(src), (void *)(dest), (n))
-#endif
 
 #define CHAR_ALPHA		'a' /* Alphabetic characters only */
 #define CHAR_BOTH		'b' /* Both - either character or digit */
@@ -75,7 +65,7 @@
 				 
 /* Declaration of Actions */
 
-/* ARGSUSED */
+
 static void ToggleOverwriteACT(Widget, XEvent *,
 				  String *, Cardinal *nparams);
 
@@ -88,7 +78,7 @@ static void Destroy(Widget);
 
 /* Utility routines */
 static void parsePattern(XbaeInputWidget, String);
-static int match(XbaeInputWidget, char *, int);
+static Boolean match(XbaeInputWidget, char *, Boolean);
 
 /* Internal callbacks */
 static void checkInput(Widget, XtPointer, XtPointer);
@@ -120,7 +110,6 @@ static XtResource resources[] = {
       offset(input.validate_callback), XmRCallback, (XtPointer)NULL }
 };
 
-#if XmVersion > 1001
 static XmPrimitiveClassExtRec primitiveClassExtRec = {
     NULL,				/* next_extension	*/
     NULLQUARK,				/* record_type		*/
@@ -130,7 +119,6 @@ static XmPrimitiveClassExtRec primitiveClassExtRec = {
     XmInheritDisplayRectProc,		/* widget_display_rect	*/
     (XmWidgetMarginsProc)_XtInherit	/* widget_margins	*/
 };
-#endif
 
 XbaeInputClassRec xbaeInputClassRec =
 {
@@ -175,11 +163,7 @@ XbaeInputClassRec xbaeInputClassRec =
 	NULL,				/* arm_and_activate	*/
 	NULL,				/* syn resources	*/
 	0,				/* num_syn_resources	*/
-#if XmVersion <= 1001
-	NULL,				/* extension		*/
-#else
-	&primitiveClassExtRec,		/* extension		*/
-#endif
+	(XtPointer)&primitiveClassExtRec, /* extension		*/
     },
     { /* Text class */
 	NULL,				/* extension		*/
@@ -195,7 +179,7 @@ WidgetClass xbaeInputWidgetClass = (WidgetClass) &xbaeInputClassRec;
  * Action routines
  */
 
-/*ARGSUSED*/
+
 static void
 validate(Widget w, XtPointer cd, XtPointer cb)
 {
@@ -205,14 +189,17 @@ validate(Widget w, XtPointer cd, XtPointer cb)
     XbaeInputValidateCallbackStruct call_data;
     String pattern = NULL;
     String value;
-    int ret;
+    Boolean ret;
 
-    if (!iw->input.pattern)
+    if (!iw->input.pattern) {
+	DEBUGOUT(_XbaeDebug(__FILE__, w, "Nothing to validate\n"));
 	return;
+    }
+    DEBUGOUT(_XbaeDebug(__FILE__, w, "Validate ...\n"));
 
     value = XmTextGetString(w);
 
-    ret = match(iw, value, (int)True);
+    ret = match(iw, value, True);
 
     if (iw->input.auto_fill && iw->input.literal_pending)
     {
@@ -263,6 +250,7 @@ validate(Widget w, XtPointer cd, XtPointer cb)
     if (!call_data.doit)
     {
 	XBell(XtDisplay(w), 0);
+	DEBUGOUT(_XbaeDebug(__FILE__, w, "ring ring\n"));
 	XmProcessTraversal(w, XmTRAVERSE_CURRENT);
 
 	if (cbs->reason == XmCR_ACTIVATE)
@@ -291,15 +279,13 @@ validate(Widget w, XtPointer cd, XtPointer cb)
 	XtFree(value);
 }
 
-/* ARGSUSED */
+
 static void
 ToggleOverwriteACT(Widget w, XEvent *event, String *params, Cardinal *nparams)
 {
     XbaeInputWidget iw = (XbaeInputWidget)w;
 
-#ifdef DEBUG
-    printf ("Toggle\n");
-#endif
+    DEBUGOUT(_XbaeDebug(__FILE__, w, "Toggle\n"));
 
     if (iw->input.overwrite_mode)
 	iw->input.overwrite_mode = False;
@@ -344,8 +330,10 @@ Initialize(Widget request, Widget new, ArgList args, Cardinal *num_args)
      * By default, overwrite mode is off on the text field.  By calling
      * the text widget's action routine it will be turned on.
      */
-    if (iw->input.overwrite_mode == True)
+    if (iw->input.overwrite_mode == True) {
+	DEBUGOUT(_XbaeDebug(__FILE__, new, "Initialize: toggle-overstrike\n"));
 	XtCallActionProc(new, "toggle-overstrike", NULL, NULL, 0);
+    }
 
     XtAddCallback(new, XmNlosingFocusCallback, validate, NULL);
     XtAddCallback(new, XmNactivateCallback, validate, NULL);
@@ -396,7 +384,7 @@ SetValues(Widget current, Widget request, Widget new,
 	if (text && *text)
 	{
 	    *(text + pos) = 0;
-	    (void)match(n, text, (int)False);
+	    match(n, text, False);
 	}
 	else
 	    parsePattern(n, new(pattern));
@@ -469,7 +457,7 @@ parsePattern(XbaeInputWidget iw, String pattern)
 	     * If the first character is a literal, set literal_pending
 	     * to True
 	     */
-	    if (iw->text.cursor_position == 0 &&
+	    if (ptr == pattern &&
 	       	iw->input.pattern_length == 0 && IS_LITERAL(*ptr))
 	    {
 		char *orig = ptr;
@@ -498,25 +486,25 @@ parsePattern(XbaeInputWidget iw, String pattern)
     XtVaSetValues((Widget)iw, XmNmaxLength, iw->input.pattern_length, NULL);
 }
 
-static int
-match(XbaeInputWidget iw, char *string, int exact)
+
+static Boolean
+match(XbaeInputWidget iw, char *string, Boolean exact)
 {
     char *tptr, *sptr, *ntptr;
     char *lastopt = NULL;	/* last optional block */
     char *lastreq = NULL;	/* last required block */
     int optional = 0;
     /*int pend_rew = 0;*/
-    int equal = 1;
+    Boolean equal = True;
     int escaped;
     int set_lastopt = 0;
 
     tptr = iw->input.pattern;
     sptr = string;
 
-#ifdef DEBUG    
-    printf("checking '%s' against '%s'\n", string ? string : "(null)",
-	   iw->input.pattern ? iw->input.pattern : "(null)");
-#endif
+    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)iw,
+	"checking '%s' against '%s'\n", string ? string : "(null)",
+	iw->input.pattern ? iw->input.pattern : "(null)"));
 
     if (!tptr || !sptr || !*tptr || !*sptr)
     {
@@ -574,13 +562,13 @@ match(XbaeInputWidget iw, char *string, int exact)
 		    set_lastopt = 0;
 		    break;
 		case CHAR_ALPHA:	/* alphabetic */
-		    equal = (int)isalpha(*sptr);
+		    equal = (isalpha(*sptr)==0) ? False : True;
 		    if (equal && set_lastopt)
 			lastopt = sptr;
 		    set_lastopt = 0;
 		    break;
 		case CHAR_BOTH:		/* both: alphabetic or digit */
-		    equal = (int)isalnum(*sptr);
+		    equal = (isalnum(*sptr)==0) ? False : True;
 		    if (equal && set_lastopt)
 			lastopt = sptr;
 		    set_lastopt = 0;
@@ -592,7 +580,7 @@ match(XbaeInputWidget iw, char *string, int exact)
 		    set_lastopt = 0;
 		    break;
 		case CHAR_DIGIT:	/* any digit */
-		    equal = (int)isdigit(*sptr);
+		    equal = (isdigit(*sptr)==0) ? False : True;
 		    if (equal && set_lastopt)
 			lastopt = sptr;
 		    set_lastopt = 0;
@@ -600,7 +588,7 @@ match(XbaeInputWidget iw, char *string, int exact)
 		case CHAR_UPPER:
 		    if (iw->input.convert_case && islower(*sptr))
 			*sptr = toupper(*sptr);
-		    equal = (int)(isalpha(*sptr) && isupper(*sptr));
+		    equal = ((isalpha(*sptr)!=0) && isupper(*sptr));
 		    if (equal && set_lastopt)
 			lastopt = sptr;
 		    set_lastopt = 0;
@@ -608,7 +596,7 @@ match(XbaeInputWidget iw, char *string, int exact)
 		case CHAR_LOWER:
 		    if (iw->input.convert_case && isupper(*sptr))
 			*sptr = tolower(*sptr);
-		    equal = (int)(isalpha(*sptr) && islower(*sptr));
+		    equal = ((isalpha(*sptr)!=0) && islower(*sptr));
 		    if (equal && set_lastopt)
 			lastopt = sptr;
 		    set_lastopt = 0;
@@ -618,8 +606,10 @@ match(XbaeInputWidget iw, char *string, int exact)
 		       If tptr is | then we already had success on the previous
 		       char so skip it and then next.
 		    */
+		    if (*ntptr == CHAR_ESCAPE)
+			ntptr++;
 		    tptr = ntptr + 1;
-		    equal = 0;
+		    /*equal = False;*/
 		    continue;
 		case CHAR_ESCAPE:
 		    tptr++;
@@ -627,7 +617,7 @@ match(XbaeInputWidget iw, char *string, int exact)
 		    escaped = 1;
 		    /* And fall through to literal */
 		default:		/* treat as a literal */
-		    equal = (int)(*sptr == *tptr);
+		    equal = (*sptr == *tptr);
 		    if (equal && set_lastopt)
 			lastopt = sptr;
 		    set_lastopt = 0;
@@ -676,10 +666,14 @@ match(XbaeInputWidget iw, char *string, int exact)
 		while (*tptr &&
 		       (*tptr != CHAR_EOPTION || *(tptr-1) == CHAR_ESCAPE))
 		    tptr++;
-		if (!*tptr)
-		    (void)fprintf(stderr, "Mismatched parentheses\n");
-		else
-		{
+		if (!*tptr) {
+		    XtAppWarningMsg(XtWidgetToApplicationContext((Widget)iw),
+			"XbaeInput",
+			"InputParenthesisMismatch",
+			"",
+			"Mismatched parentheses",
+			NULL, 0);
+		} else {
 		    tptr++;	/* move past the ']' */
 #ifdef OLD
 		    if (!lastopt)
@@ -792,7 +786,8 @@ match(XbaeInputWidget iw, char *string, int exact)
 		}
 	    }
 	    continue;
-	}
+	} /* if (!equal) */
+
 	if (escaped || (*tptr != CHAR_EOPTION && *tptr != CHAR_SOPTION))
 	{
 #ifdef DEBUG
@@ -812,8 +807,20 @@ match(XbaeInputWidget iw, char *string, int exact)
 #endif
     }
 
-    if (optional && *tptr && *tptr == CHAR_EOPTION) /* step over last ']' */
+    /* skip over the remain optional part which has been matched */
+    /* XXX don't know whether optional blocks should be treated need to
+       match entirely if exact is true. If this is the case, we should only
+       skip over CHAR_OR inside the optional. This code was more complicated
+       however, so I didn't bother. -cg
+    */
+    while (optional && *tptr)
+    {
+	if (*tptr == CHAR_ESCAPE)
 	tptr++;
+	else if (*tptr == CHAR_EOPTION)
+	    optional = 0;
+	tptr++;
+    }
 
     /*
      * Determine whether any (and how many) literals are about to be inserted
@@ -867,7 +874,7 @@ match(XbaeInputWidget iw, char *string, int exact)
     return equal;
 }
 
-/*ARGSUSED*/
+
 static void
 checkInput(Widget w, XtPointer cd, XtPointer cb)
 {
@@ -880,8 +887,11 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
     /*
      * Ignore events generated by convenience routines
      */
-    if (!cbs->event)
+    if (!cbs->event) {
+	DEBUGOUT(_XbaeDebug(__FILE__, w, "checkInput: no event\n"));
 	return;
+    }
+
     /*
      * Was it a backspace?
      */
@@ -896,14 +906,14 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
 	 * to move the cursor to overwrite the text, not to delete the character
 	 * then inserting a new value */
 	if (iw->input.overwrite_mode == True) {
-		fprintf(stderr, "Yow\n");
 		XmTextSetInsertionPosition((Widget)iw,(cbs->currInsert - 1));
 		cbs->doit = False;
 	}
 #endif
 	/*
 	 * Allow a backspace or delete to do it's bit.
-	 */	
+	 */
+	DEBUGOUT(_XbaeDebug(__FILE__, w, "CheckInput backspace\n"));
 	return;
     }
     
@@ -915,6 +925,7 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
     if (cbs->text->length > 1)
     {
 	cbs->doit = False;
+	DEBUGOUT(_XbaeDebug(__FILE__, w, "CheckInput paste ?\n"));
 	return;
     }
 
@@ -924,6 +935,7 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
     if (cbs->text->length + cbs->startPos > iw->input.pattern_length)
     {
 	cbs->doit = False;
+	DEBUGOUT(_XbaeDebug(__FILE__, w, "CheckInput too long\n"));
 	return;
     }
 
@@ -937,6 +949,7 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
     text = XmTextGetString(w);    
     pos = cbs->startPos;
 
+    DEBUGOUT(_XbaeDebug(__FILE__, w, "checkInput(%s)\n", text));
     /*
      * Copy the characters forward if inserting or at end of string.
      * Include the NUL terminator in the move.
@@ -958,11 +971,11 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
      * any pending literals may also have changed
      */
     if (iw->input.last_insert != pos ||
-	(pos < strlen(text) && iw->input.overwrite_mode == True))
+	(pos < (int)strlen(text) && iw->input.overwrite_mode == True))
     {
 	char *buf = XtNewString(text);
 	buf[pos] = 0;
-	(void)match(iw, buf, (int)False);
+	match(iw, buf, False);
 	XtFree(buf);
     }
 
@@ -984,7 +997,7 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
 		 * If in insert mode or at the end of the string,
 		 * extra space needs to be added
 		 */
-		if (!iw->input.overwrite_mode || pos >= strlen(text))
+		if (!iw->input.overwrite_mode || pos >= (int)strlen(text))
 		{
 		    length++;
 		    text = XtRealloc((void *)text, length);
@@ -1017,9 +1030,11 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
 
 	text[pos] = cbs->text->ptr[i];
 
-	if (match(iw, text, (int)False) == 0)
+	if (!match(iw, text, False))
 	{
 	    cbs->doit = False;
+	    DEBUGOUT(_XbaeDebug(__FILE__, w, "checkInput no match(%s) pat [%s]\n",
+		text, iw->input.pattern));
 	    XtFree(text);
 	    return;
 	}
@@ -1032,8 +1047,10 @@ checkInput(Widget w, XtPointer cd, XtPointer cb)
     iw->input.last_insert = pos;
 
     XtFree(text);
+    DEBUGOUT(_XbaeDebug(__FILE__, w, "checkInput return\n"));
 }
 
+
 Widget
 XbaeCreateInput(Widget parent, String name, ArgList args, Cardinal ac) 
 { 
