@@ -1,6 +1,8 @@
 /*
  * Copyright(c) 1992 Bell Communications Research, Inc. (Bellcore)
  * Copyright(c) 1995-99 Andrew Lister
+ * Copyright  1999, 2000, 2001, 2002, 2003 by the LessTif Developers.
+ *
  *                        All rights reserved
  * Permission to use, copy, modify and distribute this material for
  * any purpose and without fee is hereby granted, provided that the
@@ -20,7 +22,7 @@
  * LOST PROFITS OR OTHER INCIDENTAL OR CONSEQUENTIAL DAMAGES RELAT-
  * ING TO THE SOFTWARE.
  *
- * $Id: Shadow.c,v 1.3 2000/03/28 21:44:11 danny Exp $
+ * $Id: Shadow.c,v 1.19 2003/05/24 15:06:28 dannybackx Exp $
  */
 
 
@@ -33,16 +35,19 @@
 #include <XbaeConfig.h>
 #endif
 
+#include <stdio.h>
+
 #include <Xm/Xm.h>
 #include <Xm/XmP.h>
-#if XmVersion > 1001
 #include <Xm/DrawP.h>
-#endif
 #include <Xbae/MatrixP.h>
 #include <Xbae/Shadow.h>
 #include <Xbae/Draw.h>
 #include <Xbae/Utils.h>
 
+#include <XbaeDebug.h>
+
+
 static void DrawRowShadow(XbaeMatrixWidget, Window, int, int, int,
 			     int, int, int, GC, GC);
 
@@ -61,7 +66,14 @@ xbaeDrawCellShadow(XbaeMatrixWidget mw, Window win, int row, int column, int x,
 {
     unsigned char grid_type;
     unsigned char shadow;
-    
+
+#if 1
+/* amai, that's too verbose ...*/
+    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)mw,
+	"xbaeDrawCellShadow [%d,%d] wid %d, ht %d\n",
+	row, column, width, height));
+#endif    
+
     if ((mw->matrix.cell_shadow_thickness == 0) &&
 	(!IN_GRID_ROW_MODE(mw)) && (!IN_GRID_COLUMN_MODE(mw)))
 	return;
@@ -69,16 +81,19 @@ xbaeDrawCellShadow(XbaeMatrixWidget mw, Window win, int row, int column, int x,
     /*
      * Surround the cell with a shadow.
      */
-    if(label)
-    {
+    if (label) {
 	shadow = pressed ? XmSHADOW_IN : XmSHADOW_OUT;
 	grid_type = XmGRID_CELL_SHADOW;
-    }
-    else
-    {
+    } else {
+#ifdef	XBAE_PER_CELL
+	shadow = mw->matrix.per_cell ?
+	    mw->matrix.per_cell[row][column].shadow_type :
+	    mw->matrix.cell_shadow_type;
+#else
 	shadow = mw->matrix.cell_shadow_types ?
 	    mw->matrix.cell_shadow_types[row][column] :
 	    mw->matrix.cell_shadow_type;
+#endif
 	grid_type = mw->matrix.grid_type;
     }
 	
@@ -176,7 +191,7 @@ xbaeDrawCellShadow(XbaeMatrixWidget mw, Window win, int row, int column, int x,
     }    
 }
 
-#if XmVersion >= 1002
+
 void
 xbaeDrawCellHighlight(XbaeMatrixWidget mw,
 	Window win,
@@ -190,28 +205,40 @@ xbaeDrawCellHighlight(XbaeMatrixWidget mw,
 	int reason)
 {    
     int thick;
+	Boolean	hl;
 
+    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)mw,
+	"xbaeDrawCellHighlight [%d,%d], wid %d, ht %d\n",
+	row, column, width, height));
+
+#ifdef	XBAE_PER_CELL
+    if (!mw->matrix.per_cell) /* Just a precaution */
+	return;
+    if (!mw->matrix.per_cell[row][column].highlighted)
+	return;			/* Nothing to do! */
+    hl = mw->matrix.per_cell[row][column].highlighted;
+#else
     if (!mw->matrix.highlighted_cells) /* Just a precaution */
 	return;
-
     if (!mw->matrix.highlighted_cells[row][column])
 	return;			/* Nothing to do! */
+    hl = mw->matrix.highlighted_cells[row][column];
+#endif
 
-    if (reason & HIGHLIGHTING_SOMETHING)
+    if (reason & HIGHLIGHTING_SOMETHING) {
 	gc = mw->manager.highlight_GC;
+    }
 
     if (IN_GRID_ROW_MODE(mw) &&
 	(reason & HighlightRow || reason & UnhighlightRow) &&
-	mw->matrix.highlighted_cells[row][column] == HighlightRow)
-	DrawRowHighlight(mw, win, gc, row, column, x, y, width, height,
-			 reason);
-    else if (IN_GRID_COLUMN_MODE(mw) &&
-	     (reason & HighlightColumn || reason & UnhighlightColumn) &&
-	mw->matrix.highlighted_cells[row][column] == HighlightColumn)
+	hl == HighlightRow) {
+	DrawRowHighlight(mw, win, gc, row, column, x, y, width, height, reason);
+    } else if (IN_GRID_COLUMN_MODE(mw) &&
+		(reason & HighlightColumn || reason & UnhighlightColumn) &&
+		hl == HighlightColumn) {
 	DrawColumnHighlight(mw, win, gc, row, column, x, y, width, height,
 			    reason);
-    else
-    {
+    } else {
 	thick = (2 * mw->matrix.cell_shadow_thickness);
 
 	DRAW_HIGHLIGHT(XtDisplay(mw), win, gc,
@@ -236,7 +263,11 @@ DrawRowHighlight(XbaeMatrixWidget mw,
 	int reason)
 {
     XRectangle rect[1];
-    
+
+    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)mw,
+	"DrawRowHighlight [%d,%d] wid %d ht %d\n",
+	row, column, width, height));
+
     rect[0].x = 0;
     rect[0].y = 0;
     rect[0].width = width;
@@ -273,19 +304,38 @@ DrawRowHighlight(XbaeMatrixWidget mw,
 
 	if (NEED_HORIZ_FILL(mw))
 	{
-	    int ax, ay;
+	    int ax, ay, aay;
 
 	    xbaeCalcHorizFill(mw, win, x, y, row, column, &ax, &ay,
 			      &width, &height);
+
 	    rect[0].width = width;
 	    rect[0].height = height;
-	    XSetClipRectangles(XtDisplay(mw), gc, ax, ay, rect, 1,
+	    /*
+	     * If we're above the row label offset, only draw part of
+	     * the highlight.
+	     * As in Draw.c, xbaeCalcHorizFill() cannot be expected to
+	     * handle all this.
+	     */
+	    aay = ay;
+	    if (ay < ROW_LABEL_OFFSET(mw)) {
+		rect[0].height -= ay
+			+ 2 * mw->matrix.cell_shadow_thickness;
+		aay = ROW_LABEL_OFFSET(mw);
+	    }
+	    XSetClipRectangles(XtDisplay(mw), gc, ax, aay, rect, 1,
 			       Unsorted);
+
 	    height -= mw->matrix.cell_shadow_thickness * 2;
 	    ax -= mw->matrix.cell_highlight_thickness;
 	    width += (mw->matrix.cell_highlight_thickness -
 		      mw->matrix.cell_shadow_thickness);
 		
+	    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)mw,
+	    	"DrawRowHighlight ax %d ay %d wid %d ht %d - rect x %d y %d w %d h %d\n",
+		ax, ay, width, height,
+		rect[0].x, rect[0].y, rect[0].width, rect[0].height));
+
 	    DRAW_HIGHLIGHT(XtDisplay(mw), XtWindow(mw), gc, ax, ay,
 			    width, height,
 			    mw->matrix.cell_highlight_thickness,
@@ -397,7 +447,6 @@ DrawColumnHighlight(XbaeMatrixWidget mw,
     XSetClipMask(XtDisplay(mw), gc, None);
 }
 
-#endif
 
 static void
 DrawRowShadow(XbaeMatrixWidget mw,
@@ -415,6 +464,10 @@ DrawRowShadow(XbaeMatrixWidget mw,
     unsigned char shadow = mw->matrix.row_shadow_types ?
 	mw->matrix.row_shadow_types[row] : mw->matrix.cell_shadow_type;
 
+    DEBUGOUT(_XbaeDebug(__FILE__, (Widget)mw,
+	"DrawRowShadow [%d,%d] wid %d, ht %d\n",
+	row, column, width, height));
+
     rect[0].x = 0;
     rect[0].y = 0;
     rect[0].width = width;
@@ -494,7 +547,7 @@ DrawRowShadow(XbaeMatrixWidget mw,
 	     */
 
 	    if ((win == XtWindow(ClipChild(mw))) &&
-		(height != ROW_HEIGHT(mw)))
+		(height != SOME_ROW_HEIGHT(mw, row)))
 	    {
 		if (height == (ClipChild(mw)->core.height +
 			       ClipChild(mw)->core.y - ay))
@@ -513,7 +566,7 @@ DrawRowShadow(XbaeMatrixWidget mw,
 	     * we've got trailing fixed columns. */
 	    if (mw->matrix.trailing_fixed_columns &&
 		(win == XtWindow(RightClip(mw))) &&
-		(height != ROW_HEIGHT(mw)))
+		(height != SOME_ROW_HEIGHT(mw, row)))
 	    {
 		if (height == (RightClip(mw)->core.height +
 			       RightClip(mw)->core.y - ay))
